# -*- coding:utf-8 -*-
import unittest

from tyme4py.eightchar import EightChar, ChildLimit, Fortune, DecadeFortune
from tyme4py.eightchar.provider.impl import China95ChildLimitProvider, LunarSect2EightCharProvider, DefaultEightCharProvider, DefaultChildLimitProvider, LunarSect1ChildLimitProvider, LunarSect2ChildLimitProvider
from tyme4py.enums import Gender
from tyme4py.lunar import LunarHour
from tyme4py.sixtycycle import SixtyCycle, HeavenStem
from tyme4py.solar import SolarTime


class TestEightChar(unittest.TestCase):

    def test1(self):
        """十神"""
        # 八字
        eight_char: EightChar = EightChar(SixtyCycle('丙寅'), SixtyCycle('癸巳'), SixtyCycle('癸酉'), SixtyCycle('己未'))
        # 年柱
        year: SixtyCycle = eight_char.get_year()
        # 月柱
        month: SixtyCycle = eight_char.get_month()
        # 日柱
        day: SixtyCycle = eight_char.get_day()
        # 时柱
        hour: SixtyCycle = eight_char.get_hour()
        # 日元(日主、日干)
        me: HeavenStem = day.get_heaven_stem()
        # 年柱天干十神
        assert me.get_ten_star(year.get_heaven_stem()).get_name() == '正财'
        # 月柱天干十神
        assert me.get_ten_star(month.get_heaven_stem()).get_name() == '比肩'
        # 时柱天干十神
        assert me.get_ten_star(hour.get_heaven_stem()).get_name() == '七杀'
        # 年柱地支十神（本气)
        assert me.get_ten_star(year.get_earth_branch().get_hide_heaven_stem_main()).get_name() == '伤官'
        # 年柱地支十神（中气)
        assert year.get_earth_branch().get_hide_heaven_stem_middle()
        assert me.get_ten_star(year.get_earth_branch().get_hide_heaven_stem_middle()).get_name() == '正财'
        # 年柱地支十神（余气)
        assert year.get_earth_branch().get_hide_heaven_stem_residual()
        assert me.get_ten_star(year.get_earth_branch().get_hide_heaven_stem_residual()).get_name() == '正官'
        # 日柱地支十神（本气)
        assert me.get_ten_star(day.get_earth_branch().get_hide_heaven_stem_main()).get_name() == '偏印'
        # 日柱地支藏干（中气)
        assert day.get_earth_branch().get_hide_heaven_stem_middle() is None
        # 日柱地支藏干（余气)
        assert day.get_earth_branch().get_hide_heaven_stem_residual() is None
        # 指定任意天干的十神
        assert me.get_ten_star(HeavenStem('丙')).get_name() == '正财'

    def test2(self):
        """地势(长生十二神)"""
        # 八字
        eight_char: EightChar = EightChar(SixtyCycle('丙寅'), SixtyCycle('癸巳'), SixtyCycle('癸酉'), SixtyCycle('己未'))
        # 年柱
        year: SixtyCycle = eight_char.get_year()
        # 月柱
        month: SixtyCycle = eight_char.get_month()
        # 日柱
        day: SixtyCycle = eight_char.get_day()
        # 时柱
        hour: SixtyCycle = eight_char.get_hour()
        # 日元(日主、日干)
        me: HeavenStem = day.get_heaven_stem()
        # 年柱地势
        assert me.get_terrain(year.get_earth_branch()).get_name() == '沐浴'
        # 月柱地势
        assert me.get_terrain(month.get_earth_branch()).get_name() == '胎'
        # 日柱地势
        assert me.get_terrain(day.get_earth_branch()).get_name() == '病'
        # 时柱地势
        assert me.get_terrain(hour.get_earth_branch()).get_name() == '墓'

    def test3(self):
        """胎元 / 胎息 / 命宫"""
        # 八字
        eight_char: EightChar = EightChar(SixtyCycle('癸卯'), SixtyCycle('辛酉'), SixtyCycle('己亥'), SixtyCycle('癸酉'))
        # 胎元
        tai_yuan: SixtyCycle = eight_char.get_fetal_origin()
        assert tai_yuan.get_name() == '壬子'
        # 胎元纳音
        assert tai_yuan.get_sound().get_name() == '桑柘木'

    def test4(self):
        """胎息"""
        # 八字
        eight_char: EightChar = EightChar(SixtyCycle('癸卯'), SixtyCycle('辛酉'), SixtyCycle('己亥'), SixtyCycle('癸酉'))
        # 胎息
        tai_xi: SixtyCycle = eight_char.get_fetal_breath()
        assert tai_xi.get_name() == '甲寅'
        # 胎息纳音
        assert tai_xi.get_sound().get_name() == '大溪水'

    def test5(self):
        """命宫"""
        # 八字
        eight_char: EightChar = EightChar(SixtyCycle('癸卯'), SixtyCycle('辛酉'), SixtyCycle('己亥'), SixtyCycle('癸酉'))
        # 命宫
        ming_gong: SixtyCycle = eight_char.get_own_sign()
        assert ming_gong.get_name() == '癸亥'
        # 命宫纳音
        assert ming_gong.get_sound().get_name() == '大海水'

    def test6(self):
        """身宫"""
        # 八字
        eight_char: EightChar = EightChar(SixtyCycle('癸卯'), SixtyCycle('辛酉'), SixtyCycle('己亥'), SixtyCycle('癸酉'))
        # 身宫
        shen_gong: SixtyCycle = eight_char.get_body_sign()
        assert shen_gong.get_name() == '己未'
        # 身宫纳音
        assert shen_gong.get_sound().get_name() == '天上火'

    def test7(self):
        """地势(长生十二神)"""
        # 八字
        eight_char: EightChar = EightChar(SixtyCycle('乙酉'), SixtyCycle('戊子'), SixtyCycle('辛巳'), SixtyCycle('壬辰'))
        # 日干
        me: HeavenStem = eight_char.get_day().get_heaven_stem()
        # 年柱地势
        assert me.get_terrain(eight_char.get_year().get_earth_branch()).get_name() == '临官'
        # 月柱地势
        assert me.get_terrain(eight_char.get_month().get_earth_branch()).get_name() == '长生'
        # 日柱地势
        assert me.get_terrain(eight_char.get_day().get_earth_branch()).get_name() == '死'
        # 时柱地势
        assert me.get_terrain(eight_char.get_hour().get_earth_branch()).get_name() == '墓'

    def test8(self):
        """公历时刻转八字"""
        eight_char: EightChar = SolarTime(2005, 12, 23, 8, 37, 0).get_lunar_hour().get_eight_char()
        assert eight_char.get_year().get_name() == '乙酉'
        assert eight_char.get_month().get_name() == '戊子'
        assert eight_char.get_day().get_name() == '辛巳'
        assert eight_char.get_hour().get_name() == '壬辰'

    def test9(self):
        eight_char: EightChar = SolarTime(1988, 2, 15, 23, 30, 0).get_lunar_hour().get_eight_char()
        assert eight_char.get_year().get_name() == '戊辰'
        assert eight_char.get_month().get_name() == '甲寅'
        assert eight_char.get_day().get_name() == '辛丑'
        assert eight_char.get_hour().get_name() == '戊子'

    def test11(self):
        """童限测试"""
        child_limit: ChildLimit = ChildLimit(SolarTime(2022, 3, 9, 20, 51, 0), Gender.MAN)
        assert child_limit.get_year_count() == 8
        assert child_limit.get_month_count() == 9
        assert child_limit.get_day_count() == 2
        assert child_limit.get_hour_count() == 10
        assert child_limit.get_minute_count() == 28
        assert child_limit.get_end_time().__str__() == '2030年12月12日 07:19:00'

    def test12(self):
        """童限测试"""
        child_limit: ChildLimit = ChildLimit(SolarTime(2018, 6, 11, 9, 30, 0), Gender.WOMAN)
        assert child_limit.get_year_count() == 1
        assert child_limit.get_month_count() == 9
        assert child_limit.get_day_count() == 10
        assert child_limit.get_hour_count() == 1
        assert child_limit.get_minute_count() == 42
        assert child_limit.get_end_time().__str__() == '2020年3月21日 11:12:00'

    def test13(self):
        """大运测试"""
        # 童限
        child_limit: ChildLimit = ChildLimit(SolarTime(1983, 2, 15, 20, 0, 0), Gender.WOMAN)
        # 八字
        assert child_limit.get_eight_char().__str__() == '癸亥 甲寅 甲戌 甲戌'
        # 童限年数
        assert child_limit.get_year_count() == 6
        # 童限月数
        assert child_limit.get_month_count() == 2
        # 童限日数
        assert child_limit.get_day_count() == 18
        # 童限结束(即开始起运)的公历时刻
        assert child_limit.get_end_time().__str__() == '1989年5月4日 18:24:00'
        # 童限开始(即出生)的农历年干支
        assert child_limit.get_start_time().get_lunar_hour().get_lunar_day().get_lunar_month().get_lunar_year().get_sixty_cycle().get_name() == '癸亥'
        # 童限结束(即开始起运)的农历年干支
        assert child_limit.get_end_time().get_lunar_hour().get_lunar_day().get_lunar_month().get_lunar_year().get_sixty_cycle().get_name() == '己巳'
        # 第1轮大运
        decadeFortune: DecadeFortune = child_limit.get_start_decade_fortune()
        # 开始年龄
        assert decadeFortune.get_start_age() == 7
        # 结束年龄
        assert decadeFortune.get_end_age() == 16
        # 开始年
        assert decadeFortune.get_start_sixty_cycle_year().get_year() == 1989
        # 结束年
        assert decadeFortune.get_end_sixty_cycle_year().get_year() == 1998
        # 干支
        assert decadeFortune.get_name() == '乙卯'
        # 下一大运
        assert decadeFortune.next(1).get_name() == '丙辰'
        # 上一大运
        assert decadeFortune.next(-1).get_name() == '甲寅'
        # 第9轮大运
        assert decadeFortune.next(8).get_name() == '癸亥'
        # 小运
        fortune: Fortune = child_limit.get_start_fortune()
        # 年龄
        assert fortune.get_age() == 7
        # 农历年
        assert fortune.get_sixty_cycle_year().get_year() == 1989
        # 干支
        assert fortune.get_name() == '辛巳'
        # 流年
        assert fortune.get_sixty_cycle_year().get_sixty_cycle().get_name() == '己巳'

    def test14(self):
        # 童限
        child_limit: ChildLimit = ChildLimit(SolarTime(1992, 2, 2, 12, 0, 0), Gender.MAN)
        # 八字
        assert child_limit.get_eight_char().__str__() == '辛未 辛丑 戊申 戊午'
        # 童限年数
        assert child_limit.get_year_count() == 9
        # 童限月数
        assert child_limit.get_month_count() == 0
        # 童限日数
        assert child_limit.get_day_count() == 9
        # 童限结束(即开始起运)的公历时刻
        assert child_limit.get_end_time().__str__() == '2001年2月11日 18:58:00'
        # 童限开始(即出生)的农历年干支
        assert child_limit.get_start_time().get_lunar_hour().get_lunar_day().get_lunar_month().get_lunar_year().get_sixty_cycle().get_name() == '辛未'
        # 童限结束(即开始起运)的农历年干支
        assert child_limit.get_end_time().get_lunar_hour().get_lunar_day().get_lunar_month().get_lunar_year().get_sixty_cycle().get_name() == '辛巳'
        # 第1轮大运
        decadeFortune: DecadeFortune = child_limit.get_start_decade_fortune()
        # 开始年龄
        assert decadeFortune.get_start_age() == 10
        # 结束年龄
        assert decadeFortune.get_end_age() == 19
        # 开始年
        assert decadeFortune.get_start_sixty_cycle_year().get_year() == 2001
        # 结束年
        assert decadeFortune.get_end_sixty_cycle_year().get_year() == 2010
        # 干支
        assert decadeFortune.get_name() == '庚子'
        # 下一大运
        assert decadeFortune.next(1).get_name() == '己亥'
        # 小运
        fortune: Fortune = child_limit.get_start_fortune()
        # 年龄
        assert fortune.get_age() == 10
        # 农历年
        assert fortune.get_sixty_cycle_year().get_year() == 2001
        # 干支
        assert fortune.get_name() == '戊申'
        # 小运推移
        assert fortune.next(2).get_name() == '丙午'
        assert fortune.next(-2).get_name() == '庚戌'
        # 流年
        assert fortune.get_sixty_cycle_year().get_sixty_cycle().get_name() == '辛巳'

    def test15(self):
        assert SolarTime(2018, 8, 8, 8, 8, 0).get_lunar_hour().get_eight_char().__str__() == '戊戌 庚申 壬申 甲辰'

    def test16(self):
        # 童限
        child_limit: ChildLimit = ChildLimit(SolarTime(1990, 3, 15, 10, 30, 0), Gender.MAN)
        # 八字
        assert child_limit.get_eight_char().__str__() == '庚午 己卯 己卯 己巳'
        # 童限年数
        assert child_limit.get_year_count() == 6
        # 童限月数
        assert child_limit.get_month_count() == 11
        # 童限日数
        assert child_limit.get_day_count() == 23
        # 童限结束(即开始起运)的公历时刻
        assert child_limit.get_end_time().__str__() == '1997年3月11日 00:22:00'
        # 小运
        fortune: Fortune = child_limit.get_start_fortune()
        # 年龄
        assert fortune.get_age() == 8

    def test17(self):
        eight_char: EightChar = EightChar(SixtyCycle('己丑'), SixtyCycle('戊辰'), SixtyCycle('戊辰'), SixtyCycle('甲子'))
        assert eight_char.get_own_sign().get_name() == '丁丑'

    def test18(self):
        eight_char: EightChar = EightChar(SixtyCycle('戊戌'), SixtyCycle('庚申'), SixtyCycle('丁亥'), SixtyCycle('丙午'))
        assert eight_char.get_own_sign().get_name() == '乙卯'

    def test19(self):
        eight_char: EightChar = EightChar(SixtyCycle('甲子'), SixtyCycle('壬申'), SixtyCycle('庚子'), SixtyCycle('乙亥'))
        assert eight_char.get_own_sign().get_name() == '甲戌'

    def test20(self):
        eight_char: EightChar = ChildLimit(SolarTime(2024, 1, 29, 9, 33, 0), Gender.MAN).get_eight_char()
        assert eight_char.get_own_sign().get_name() == '癸亥'
        assert eight_char.get_body_sign().get_name() == '己未'

    def test21(self):
        eight_char: EightChar = EightChar(SixtyCycle('辛亥'), SixtyCycle('乙未'), SixtyCycle('庚子'), SixtyCycle('甲辰'))
        assert eight_char.get_body_sign().get_name() == '庚子'

    def test22(self):
        assert ChildLimit(SolarTime(1990, 1, 27, 0, 0, 0), Gender.MAN).get_eight_char().get_body_sign().get_name() == '丙寅'

    def test23(self):
        assert ChildLimit(SolarTime(2019, 3, 7, 8, 0, 0), Gender.MAN).get_eight_char().get_own_sign().get_name() == '甲戌'

    def test24(self):
        assert ChildLimit(SolarTime(2019, 3, 27, 2, 0, 0), Gender.MAN).get_eight_char().get_own_sign().get_name() == '丁丑'

    def test25(self):
        assert LunarHour(1994, 5, 20, 18, 0, 0).get_eight_char().get_own_sign().get_name() == '丙寅'

    def test26(self):
        assert SolarTime(1986, 5, 29, 13, 37, 0).get_lunar_hour().get_eight_char().get_body_sign().get_name() == '辛丑'

    def test27(self):
        assert SolarTime(1994, 12, 6, 2, 0, 0).get_lunar_hour().get_eight_char().get_body_sign().get_name() == '丁丑'

    def test28(self):
        eight_char: EightChar = EightChar(SixtyCycle('辛亥'), SixtyCycle('丁酉'), SixtyCycle('丙午'), SixtyCycle('癸巳'))
        assert eight_char.get_own_sign().get_name() == '辛卯'

    def test29(self):
        eight_char: EightChar = EightChar(SixtyCycle('丙寅'), SixtyCycle('庚寅'), SixtyCycle('辛卯'), SixtyCycle('壬辰'))
        assert eight_char.get_own_sign().get_name() == '己亥'
        assert eight_char.get_body_sign().get_name() == '乙未'

    def test30(self):
        assert EightChar('壬子', '辛亥', '壬戌', '乙巳').get_body_sign().get_name() == '乙巳'

    def test31(self):
        # 采用元亨利贞的起运算法
        ChildLimit.provider = China95ChildLimitProvider()
        # 童限
        child_limit: ChildLimit = ChildLimit(SolarTime(1986, 5, 29, 13, 37, 0), Gender.MAN)
        # 童限年数
        assert child_limit.get_year_count() == 2
        # 童限月数
        assert child_limit.get_month_count() == 7
        # 童限日数
        assert child_limit.get_day_count() == 0
        # 童限时数
        assert child_limit.get_hour_count() == 0
        # 童限分数
        assert child_limit.get_minute_count() == 0
        # 童限结束(即开始起运)的公历时刻
        assert child_limit.get_end_time().__str__() == '1988年12月29日 13:37:00'
        # 为了不影响其他测试用例，恢复默认起运算法
        ChildLimit.provider = DefaultChildLimitProvider()

    def test46(self):
        LunarHour.provider = LunarSect2EightCharProvider()
        eight_char: EightChar = EightChar(SixtyCycle('壬寅'), SixtyCycle('丙午'), SixtyCycle('己亥'), SixtyCycle('丙子'))
        time_list: [str] = []
        aa = eight_char.get_solar_times(1900, 2024)
        for time in aa:
            time_list.append(time.__str__())
        assert time_list == ['1962年6月30日 23:00:00', '2022年6月15日 23:00:00']
        LunarHour.provider = DefaultEightCharProvider()

    def test47(self):
        # 采用Lunar流派1的起运算法
        ChildLimit.provider = LunarSect1ChildLimitProvider()
        # 童限
        child_limit: ChildLimit = ChildLimit(SolarTime(1986, 5, 29, 13, 37, 0), Gender.MAN)
        # 童限年数
        assert child_limit.get_year_count() == 2
        # 童限月数
        assert child_limit.get_month_count() == 7
        # 童限日数
        assert child_limit.get_day_count() == 0
        # 童限时数
        assert child_limit.get_hour_count() == 0
        # 童限分数
        assert child_limit.get_minute_count() == 0
        # 童限结束(即开始起运)的公历时刻
        assert child_limit.get_end_time().__str__() == '1988年12月29日 13:37:00'

        # 为了不影响其他测试用例，恢复默认起运算法
        ChildLimit.provider = DefaultChildLimitProvider()

    def test48(self):
        # 采用Lunar流派2的起运算法
        ChildLimit.provider = LunarSect2ChildLimitProvider()
        # 童限
        child_limit: ChildLimit = ChildLimit(SolarTime(1986, 5, 29, 13, 37, 0), Gender.MAN)
        # 童限年数
        assert child_limit.get_year_count() == 2
        # 童限月数
        assert child_limit.get_month_count() == 7
        # 童限日数
        assert child_limit.get_day_count() == 0
        # 童限时数
        assert child_limit.get_hour_count() == 14
        # 童限分数
        assert child_limit.get_minute_count() == 0
        # 童限结束(即开始起运)的公历时刻
        assert child_limit.get_end_time().__str__() == '1988年12月30日 03:37:00'
        # 为了不影响其他测试用例，恢复默认起运算法
        ChildLimit.provider = DefaultChildLimitProvider()

    def test49(self):
        # 采用Lunar流派2的八字算法
        LunarHour.provider = LunarSect2EightCharProvider()
        # 童限
        eight_char: EightChar = LunarHour(2001, 10, 18, 18, 0, 0).get_eight_char()
        assert eight_char.get_name() == '辛巳 己亥 己亥 癸酉'
        # 为了不影响其他测试用例，恢复默认八字算法
        LunarHour.provider = DefaultEightCharProvider()

    def test50(self):
        # 采用Lunar流派2的八字算法
        LunarHour.provider = LunarSect2EightCharProvider()
        # 童限
        eight_char: EightChar = LunarHour(2001, 10, 18, 18, 0, 0).get_eight_char()
        assert eight_char.get_name() == '辛巳 己亥 己亥 癸酉'
        # 为了不影响其他测试用例，恢复默认八字算法
        LunarHour.provider = DefaultEightCharProvider()

    def test51(self):
        eight_char: EightChar = EightChar('壬申', '壬寅', '庚辰', '甲申')
        time_list: [str] = []
        aa = eight_char.get_solar_times(1801, 2099)
        for time in aa:
            time_list.append(time.__str__())
        assert time_list == ['1812年2月18日 16:00:00', '1992年3月5日 15:00:00', '2052年2月19日 16:00:00']

    def test52(self):
        assert SolarTime.from_ymd_hms(1034, 10, 2, 20, 0, 0).get_lunar_hour().get_eight_char().__str__() == '甲戌 癸酉 甲戌 甲戌'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.holiday import LegalHoliday


class TestLegalHoliday(unittest.TestCase):
    def test(self):
        d = LegalHoliday.from_ymd(2011, 5, 1)
        assert d is not None
        assert d.__str__() == '2011年5月1日 劳动节(休)'
        d1 = d.next(1)
        assert d1 is not None
        assert d1.__str__() == '2011年5月2日 劳动节(休)'
        d2 = d.next(2)
        assert d2 is not None
        assert d2.__str__() == '2011年6月4日 端午节(休)'
        d3 = d.next(-1)
        assert d3 is not None
        assert d3.__str__() == '2011年4月30日 劳动节(休)'
        d4 = d.next(-2)
        assert d4 is not None
        assert d4.__str__() == '2011年4月5日 清明节(休)'

    def test3(self):
        d = LegalHoliday.from_ymd(2001, 12, 29)
        assert d is not None
        assert d.__str__(), '2001年12月29日 元旦节(班)'
        assert d.next(-1) is None

    def test4(self):
        d = LegalHoliday.from_ymd(2022, 10, 5)
        assert d is not None
        assert d.__str__() == '2022年10月5日 国庆节(休)'
        d1 = d.next(-1)
        assert d1 is not None
        assert d1.__str__() == '2022年10月4日 国庆节(休)'
        d2 = d.next(1)
        assert d2 is not None
        assert d2.__str__() == '2022年10月6日 国庆节(休)'

    def test5(self):
        d = LegalHoliday.from_ymd(2010, 10, 1)
        assert d is not None
        assert d.__str__() == '2010年10月1日 国庆节(休)'



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from tyme4py import LoopTyme
from tyme4py.culture import Land, Animal, Luck, Zone

if TYPE_CHECKING:
    from tyme4py.culture.star.seven import SevenStar


class TwentyEightStar(LoopTyme):
    """二十八宿"""
    NAMES: List[str] = ['角', '亢', '氐', '房', '心', '尾', '箕', '斗', '牛', '女', '虚', '危', '室', '壁', '奎', '娄', '胃', '昴', '毕', '觜', '参', '井', '鬼', '柳', '星', '张', '翼', '轸']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> TwentyEightStar:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> TwentyEightStar:
        return cls(index)

    def next(self, n: int) -> TwentyEightStar:
        return TwentyEightStar(self.next_index(n))

    def get_seven_star(self) -> SevenStar:
        """
        :return: 七曜
        """
        from tyme4py.culture.star.seven import SevenStar
        return SevenStar(self.get_index() % 7 + 4)

    def get_land(self) -> Land:
        """
        :return: 九野
        """
        return Land([4, 4, 4, 2, 2, 2, 7, 7, 7, 0, 0, 0, 0, 5, 5, 5, 6, 6, 6, 1, 1, 1, 8, 8, 8, 3, 3, 3][self.get_index()])

    def get_zone(self) -> Zone:
        """
        :return: 宫
        """
        return Zone(self.get_index() // 7)

    def get_animal(self) -> Animal:
        """
        :return: 动物
        """
        return Animal(self.get_index())

    def get_luck(self) -> Luck:
        """
        :return: 吉凶
        """
        return Luck([0, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1, 1, 0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 1, 1, 1, 0, 1, 0][self.get_index()])



# -*- coding:utf-8 -*-
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from tyme4py.eightchar.provider import EightCharProvider, ChildLimitProvider

if TYPE_CHECKING:
    from tyme4py.eightchar import EightChar, ChildLimitInfo
    from tyme4py.lunar import LunarHour
    from tyme4py.solar import SolarTime, SolarTerm, SolarMonth


class DefaultEightCharProvider(EightCharProvider):
    """默认的八字计算（晚子时算第二天）"""

    def get_eight_char(self, hour: LunarHour) -> EightChar:
        return hour.get_sixty_cycle_hour().get_eight_char()


class LunarSect2EightCharProvider(EightCharProvider):
    """Lunar流派2的八字计算（晚子时日柱算当天）"""

    def get_eight_char(self, hour: LunarHour) -> EightChar:
        from tyme4py.eightchar import EightChar
        from tyme4py.sixtycycle import SixtyCycleHour
        h: SixtyCycleHour = hour.get_sixty_cycle_hour()
        return EightChar(h.get_year(), h.get_month(), hour.get_lunar_day().get_sixty_cycle(), h.get_sixty_cycle())


class AbstractChildLimitProvider(ChildLimitProvider):
    """童限计算抽象"""

    @staticmethod
    def next(birth_time: SolarTime, add_year: int, add_month: int, add_day: int, add_hour: int, add_minute: int, add_second: int) -> ChildLimitInfo:
        from tyme4py.solar import SolarTime, SolarMonth
        from tyme4py.eightchar import ChildLimitInfo
        d: int = birth_time.get_day() + add_day
        h: int = birth_time.get_hour() + add_hour
        mi: int = birth_time.get_minute() + add_minute
        s: int = birth_time.get_second() + add_second
        mi += s // 60
        s %= 60
        h += mi // 60
        mi %= 60
        d += h // 24
        h %= 24
        sm: SolarMonth = SolarMonth(birth_time.get_year() + add_year, birth_time.get_month()).next(add_month)
        dc: int = sm.get_day_count()
        while d > dc:
            d -= dc
            sm = sm.next(1)
            dc = sm.get_day_count()
        return ChildLimitInfo(birth_time, SolarTime(sm.get_year(), sm.get_month(), d, h, mi, s), add_year, add_month, add_day, add_hour, add_minute)

    @abstractmethod
    def get_info(self, birth_time: SolarTime, term: SolarTerm) -> ChildLimitInfo:
        pass


class DefaultChildLimitProvider(AbstractChildLimitProvider):
    """默认的八字计算（晚子时算第二天）"""

    def get_info(self, birth_time: SolarTime, term: SolarTerm) -> ChildLimitInfo:
        # 出生时刻和节令时刻相差的秒数
        seconds: int = abs(term.get_julian_day().get_solar_time().subtract(birth_time))
        # 3天 = 1年，3天=60*60*24*3秒=259200秒 = 1年
        year: int = seconds // 259200
        seconds %= 259200
        # 1天 = 4月，1天=60*60*24秒=86400秒 = 4月，85400秒/4=21600秒 = 1月
        month: int = seconds // 21600
        seconds %= 21600
        # 1时 = 5天，1时=60*60秒=3600秒 = 5天，3600秒/5=720秒 = 1天
        day: int = seconds // 720
        seconds %= 720
        # 1分 = 2时，60秒 = 2时，60秒/2=30秒 = 1时
        hour: int = seconds // 30
        seconds %= 30
        # 1秒 = 2分，1秒/2=0.5秒 = 1分
        minute: int = seconds * 2
        return AbstractChildLimitProvider.next(birth_time, year, month, day, hour, minute, 0)


class China95ChildLimitProvider(AbstractChildLimitProvider):
    """元亨利贞的童限计算"""

    def get_info(self, birth_time: SolarTime, term: SolarTerm) -> ChildLimitInfo:
        # 出生时刻和节令时刻相差的分钟数
        minutes: int = int(abs(term.get_julian_day().get_solar_time().subtract(birth_time)) / 60)
        year: int = minutes // 4320
        minutes %= 4320
        month: int = minutes // 360
        minutes %= 360
        day: int = minutes // 12
        return AbstractChildLimitProvider.next(birth_time, year, month, day, 0, 0, 0)


class LunarSect1ChildLimitProvider(AbstractChildLimitProvider):
    """Lunar的流派1童限计算（按天数和时辰数计算，3天1年，1天4个月，1时辰10天）"""

    def get_info(self, birth_time: SolarTime, term: SolarTerm) -> ChildLimitInfo:
        term_time: SolarTime = term.get_julian_day().get_solar_time()
        end: SolarTime = term_time
        start: SolarTime = birth_time
        if birth_time.is_after(term_time):
            end = birth_time
            start = term_time

        end_time_zhi_index: int = 11 if end.get_hour() == 23 else end.get_lunar_hour().get_index_in_day()
        start_time_zhi_index: int = 11 if start.get_hour() == 23 else start.get_lunar_hour().get_index_in_day()
        # 时辰差
        hour_diff: int = end_time_zhi_index - start_time_zhi_index
        # 天数差
        day_diff: int = end.get_solar_day().subtract(start.get_solar_day())
        if hour_diff < 0:
            hour_diff += 12
            day_diff -= 1

        month_diff: int = int(hour_diff * 10 / 30)
        month: int = day_diff * 4 + month_diff
        day: int = hour_diff * 10 - month_diff * 30
        year: int = int(month / 12)
        month = month - year * 12
        return AbstractChildLimitProvider.next(birth_time, year, month, day, 0, 0, 0)


class LunarSect2ChildLimitProvider(AbstractChildLimitProvider):
    """Lunar的流派2童限计算（按分钟数计算）"""

    def get_info(self, birth_time: SolarTime, term: SolarTerm) -> ChildLimitInfo:
        # 出生时刻和节令时刻相差的分钟数
        minutes: int = int(abs(term.get_julian_day().get_solar_time().subtract(birth_time)) / 60)
        year: int = int(minutes / 4320)
        minutes %= 4320
        month: int = int(minutes / 360)
        minutes %= 360
        day: int = int(minutes / 12)
        minutes %= 12
        hour: int = minutes * 2
        return AbstractChildLimitProvider.next(birth_time, year, month, day, hour, 0, 0)



# -*- coding:utf-8 -*-
from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Union, List


class Culture(ABC):
    """传统文化(民俗)"""

    @abstractmethod
    def get_name(self) -> str:
        """
        名称
        :return: 名称
        """
        pass


class Tyme(Culture):
    @abstractmethod
    def next(self, n: int) -> Union[Tyme, None]:
        """
        推移
        :param n:推移步数
        :return: 推移后的Tyme
        """
        pass


class AbstractCulture(Culture):
    """传统文化抽象"""

    @abstractmethod
    def get_name(self) -> str:
        pass

    def __str__(self) -> str:
        return self.get_name()

    def __eq__(self, other: Culture) -> bool:
        return other and other.__str__() == self.__str__()

    @staticmethod
    def index_of(index: int, size: int) -> int:
        """
        转换为不超范围的索引
        :param index: 索引
        :param size: 长度大小
        :return: 索引，从0开始
        """
        i: int = index % size
        return i + size if i < 0 else i


class AbstractTyme(AbstractCulture, Tyme):
    """抽象Tyme"""

    @abstractmethod
    def get_name(self) -> str:
        pass

    @abstractmethod
    def next(self, n: int) -> Union[Tyme, None]:
        pass


class AbstractCultureDay(AbstractCulture):
    """带天索引的传统文化抽象"""
    _culture: AbstractCulture
    _day_index: int

    def __init__(self, culture: AbstractCulture, day_index: int):
        self._culture = culture
        self._day_index = day_index

    def get_day_index(self) -> int:
        """
        天索引
        :return: 索引
        """
        return self._day_index

    def get_culture(self) -> AbstractCulture:
        return self._culture

    def get_name(self) -> str:
        return self._culture.get_name()

    def __str__(self) -> str:
        return f'{self.get_name()}第{self.get_day_index() + 1}天'


class LoopTyme(AbstractTyme):
    """可轮回的Tyme"""
    _names: List[str]
    """名称列表"""
    _index: int
    """索引，从0开始"""

    def __init__(self, names: List[str], index_or_name: Union[int, str]):
        """
        :param names: 名称列表
        :param index_or_name: 索引，支持负数，自动轮转; 名称
        """
        self._names = names
        self._index = self._index_of_by(index_or_name)

    def _index_of_by(self, index_or_name: Union[int, str]) -> int:
        """
        :param index_or_name: 索引或名称
        :return: 索引
        """
        if isinstance(index_or_name, int):
            return self.index_of(index_or_name, self.get_size())
        else:
            for i in range(0, self.get_size()):
                if index_or_name == self._names[i]:
                    return i
            raise ValueError(f"illegal name: {index_or_name}")

    def get_name(self) -> str:
        """
        名称
        :return: 名称
        """
        return self._names[self._index]

    def get_index(self) -> int:
        """
        索引
        :return: 索引，从0开始
        """
        return self._index

    def get_size(self) -> int:
        """
        数量
        :return: 数量
        """
        return self._names.__len__()

    def next_index(self, n: int) -> int:
        """
        推移后的索引
        :param n: 推移步数
        :return: 索引，从0开始
        """
        return self._index_of_by(self._index + n)

    def steps_to(self, target_index: int) -> int:
        """
        到目标索引的步数
        :param target_index: 目标索引
        :return: 步数
        """
        return self.index_of(target_index - self._index, self.get_size())

    @abstractmethod
    def next(self, n: int) -> Union[LoopTyme, None]:
        pass



# -*- coding:utf-8 -*-
import unittest

from tyme4py.lunar import LunarHour, LunarDay, LunarMonth, LunarYear
from tyme4py.solar import SolarDay


class TestNineStar(unittest.TestCase):
    def test0(self):
        nine_star = LunarYear(1985).get_nine_star()
        assert nine_star.get_name() == '六'
        assert nine_star.__str__() == '六白金'

    def test1(self):
        nine_star = LunarYear(2022).get_nine_star()
        assert nine_star.__str__() == '五黄土'
        assert nine_star.get_dipper().__str__() == '玉衡'

    def test2(self):
        nine_star = LunarYear(2033).get_nine_star()
        assert nine_star.__str__() == '三碧木'
        assert nine_star.get_dipper().__str__() == '天玑'

    def test3(self):
        nine_star = LunarMonth.from_ym(1985, 2).get_nine_star()
        assert nine_star.__str__() == '四绿木'
        assert nine_star.get_dipper().__str__() == '天权'

    def test4(self):
        nine_star = LunarMonth.from_ym(1985, 2).get_nine_star()
        assert nine_star.__str__() == '四绿木'
        assert nine_star.get_dipper().__str__() == '天权'

    def test5(self):
        nine_star = LunarMonth.from_ym(2022, 1).get_nine_star()
        assert nine_star.__str__() == '二黑土'
        assert nine_star.get_dipper().__str__() == '天璇'

    def test6(self):
        nine_star = LunarMonth.from_ym(2033, 1).get_nine_star()
        assert nine_star.__str__() == '五黄土'
        assert nine_star.get_dipper().__str__() == '玉衡'

    def test7(self):
        nine_star = SolarDay(1985, 2, 19).get_lunar_day().get_nine_star()
        assert nine_star.__str__() == '五黄土'
        assert nine_star.get_dipper().__str__() == '玉衡'

    def test8(self):
        nine_star = LunarDay(2022, 1, 1).get_nine_star()
        assert nine_star.__str__() == '四绿木'
        assert nine_star.get_dipper().__str__() == '天权'

    def test9(self):
        nine_star = LunarDay(2033, 1, 1).get_nine_star()
        assert nine_star.__str__() == '一白水'
        assert nine_star.get_dipper().__str__() == '天枢'

    def test10(self):
        nine_star = LunarHour(2033, 1, 1, 12, 0, 0).get_nine_star()
        assert nine_star.__str__() == '七赤金'
        assert nine_star.get_dipper().__str__() == '摇光'

    def test11(self):
        nine_star = LunarHour(2011, 5, 3, 23, 0, 0).get_nine_star()
        assert nine_star.__str__() == '七赤金'
        assert nine_star.get_dipper().__str__() == '摇光'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestDirection(unittest.TestCase):
    def test(self):
        assert SolarDay(2021, 11, 13).get_lunar_day().get_sixty_cycle().get_heaven_stem().get_mascot_direction().get_name() == '东南'

    def test1(self):
        assert SolarDay(2024, 1, 1).get_lunar_day().get_sixty_cycle().get_heaven_stem().get_mascot_direction().get_name() == '东南'

    def test2(self):
        assert SolarDay(2023, 11, 6).get_lunar_day().get_jupiter_direction().get_name() == '东'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.lunar import LunarMonth, LunarDay
from tyme4py.solar import SolarDay


class TestLunarMonth(unittest.TestCase):
    def test0(self):
        assert LunarMonth.from_ym(2359, 7).get_name() == '七月'

    def test1(self):
        """闰月"""
        assert LunarMonth.from_ym(2359, -7).get_name() == '闰七月'

    def test2(self):
        assert LunarMonth.from_ym(2023, 6).get_day_count() == 29

    def test3(self):
        assert LunarMonth.from_ym(2023, 7).get_day_count() == 30

    def test4(self):
        assert LunarMonth.from_ym(2023, 8).get_day_count() == 30

    def test5(self):
        assert LunarMonth.from_ym(2023, 9).get_day_count() == 29

    def test6(self):
        assert LunarMonth.from_ym(2023, 9).get_first_julian_day().get_solar_day().__str__() == '2023年10月15日'

    def test7(self):
        assert LunarMonth.from_ym(2023, 1).get_sixty_cycle().get_name() == '甲寅'

    def test8(self):
        assert LunarMonth.from_ym(2023, -2).get_sixty_cycle().get_name() == '乙卯'

    def test9(self):
        assert LunarMonth.from_ym(2023, 3).get_sixty_cycle().get_name() == '丙辰'

    def test10(self):
        assert LunarMonth.from_ym(2024, 1).get_sixty_cycle().get_name() == '丙寅'

    def test11(self):
        assert LunarMonth.from_ym(2023, 12).get_sixty_cycle().get_name() == '乙丑'

    def test12(self):
        assert LunarMonth.from_ym(2022, 1).get_sixty_cycle().get_name() == '壬寅'

    def test13(self):
        assert LunarMonth.from_ym(37, -12).get_name() == '闰十二月'

    def test14(self):
        assert LunarMonth.from_ym(5552, -12).get_name() == '闰十二月'

    def test15(self):
        assert LunarMonth.from_ym(2008, 11).next(1).__str__() == '农历戊子年十二月'

    def test16(self):
        assert LunarMonth.from_ym(2008, 11).next(2).__str__() == '农历己丑年正月'

    def test17(self):
        assert LunarMonth.from_ym(2008, 11).next(6).__str__() == '农历己丑年五月'

    def test18(self):
        assert LunarMonth.from_ym(2008, 11).next(7).__str__() == '农历己丑年闰五月'

    def test19(self):
        assert LunarMonth.from_ym(2008, 11).next(8).__str__() == '农历己丑年六月'

    def test20(self):
        assert LunarMonth.from_ym(2008, 11).next(15).__str__() == '农历庚寅年正月'

    def test21(self):
        assert LunarMonth.from_ym(2008, 12).next(-1).__str__() == '农历戊子年十一月'

    def test22(self):
        assert LunarMonth.from_ym(2009, 1).next(-2).__str__() == '农历戊子年十一月'

    def test23(self):
        assert LunarMonth.from_ym(2009, 5).next(-6).__str__() == '农历戊子年十一月'

    def test24(self):
        assert LunarMonth.from_ym(2009, -5).next(-7).__str__() == '农历戊子年十一月'

    def test25(self):
        assert LunarMonth.from_ym(2009, 6).next(-8).__str__() == '农历戊子年十一月'

    def test26(self):
        assert LunarMonth.from_ym(2010, 1).next(-15).__str__() == '农历戊子年十一月'

    def test27(self):
        assert LunarMonth.from_ym(2012, -4).get_day_count() == 29

    def test28(self):
        assert LunarMonth.from_ym(2023, 9).get_sixty_cycle().__str__() == '壬戌'

    def test29(self):
        d: LunarDay = SolarDay(2023, 10, 7).get_lunar_day()
        assert d.get_lunar_month().get_sixty_cycle().__str__() == '辛酉'
        assert d.get_sixty_cycle_day().get_month().__str__() == '辛酉'

    def test30(self):
        d: LunarDay = SolarDay(2023, 10, 8).get_lunar_day()
        assert d.get_lunar_month().get_sixty_cycle().__str__() == '辛酉'
        assert d.get_sixty_cycle_day().get_month().__str__() == '壬戌'

    def test31(self):
        d: LunarDay = SolarDay(2023, 10, 15).get_lunar_day()
        assert d.get_lunar_month().get_name() == '九月'
        assert d.get_lunar_month().get_sixty_cycle().__str__() == '壬戌'
        assert d.get_sixty_cycle_day().get_month().__str__() == '壬戌'

    def test32(self):
        d: LunarDay = SolarDay(2023, 11, 7).get_lunar_day()
        assert d.get_lunar_month().get_sixty_cycle().__str__() == '壬戌'
        assert d.get_sixty_cycle_day().get_month().__str__() == '壬戌'

    def test33(self):
        d: LunarDay = SolarDay(2023, 11, 8).get_lunar_day()
        assert d.get_lunar_month().get_sixty_cycle().__str__() == '壬戌'
        assert d.get_sixty_cycle_day().get_month().__str__() == '癸亥'

    def test34(self):
        # 2023年闰2月
        m: LunarMonth = LunarMonth.from_ym(2023, 12)
        assert m.__str__() == '农历癸卯年十二月'
        assert m.next(-1).__str__() == '农历癸卯年十一月'
        assert m.next(-2).__str__() == '农历癸卯年十月'

    def test35(self):
        # 2023年闰2月
        m: LunarMonth = LunarMonth.from_ym(2023, 3)
        assert m.__str__() == '农历癸卯年三月'
        assert m.next(-1).__str__() == '农历癸卯年闰二月'
        assert m.next(-2).__str__() == '农历癸卯年二月'
        assert m.next(-3).__str__() == '农历癸卯年正月'
        assert m.next(-4).__str__() == '农历壬寅年十二月'
        assert m.next(-5).__str__() == '农历壬寅年十一月'

    def test36(self):
        d: LunarDay = SolarDay(1983, 2, 15).get_lunar_day()
        assert d.get_lunar_month().get_sixty_cycle().__str__() == '甲寅'
        assert d.get_sixty_cycle_day().get_month().__str__() == '甲寅'

    def test37(self):
        d: LunarDay = SolarDay(2023, 10, 30).get_lunar_day()
        assert d.get_lunar_month().get_sixty_cycle().__str__() == '壬戌'
        assert d.get_sixty_cycle_day().get_month().__str__() == '壬戌'

    def test38(self):
        d: LunarDay = SolarDay(2023, 10, 19).get_lunar_day()
        assert d.get_lunar_month().get_sixty_cycle().__str__() == '壬戌'
        assert d.get_sixty_cycle_day().get_month().__str__() == '壬戌'

    def test39(self):
        m: LunarMonth = LunarMonth.from_ym(2023, 11)
        assert m.__str__() == '农历癸卯年十一月'
        assert m.get_sixty_cycle().__str__() == '甲子'

    def test40(self):
        assert LunarMonth.from_ym(2018, 6).get_sixty_cycle().__str__() == '己未'

    def test41(self):
        assert LunarMonth.from_ym(2017, 12).get_sixty_cycle().__str__() == '癸丑'

    def test42(self):
        assert LunarMonth.from_ym(2018, 1).get_sixty_cycle().__str__() == '甲寅'

    def test43(self):
        assert LunarDay(2018, 6, 26).get_sixty_cycle_day().get_month().__str__() == '庚申'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.lunar import LunarYear


class TestLunarYear(unittest.TestCase):
    def test0(self):
        assert LunarYear(2023).get_name() == '农历癸卯年'

    def test1(self):
        assert LunarYear(2023).next(5).get_name() == '农历戊申年'

    def test2(self):
        assert LunarYear(2023).next(-5).get_name() == '农历戊戌年'

    def test3(self):
        """农历年的干支"""
        assert LunarYear(2020).get_sixty_cycle().get_name() == '庚子'

    def test4(self):
        """农历年的生肖(农历年.干支.地支.生肖)"""
        assert LunarYear(1986).get_sixty_cycle().get_earth_branch().get_zodiac().get_name() == '虎'

    def test5(self):
        assert LunarYear(151).get_leap_month() == 12

    def test6(self):
        assert LunarYear(2357).get_leap_month() == 1

    def test7(self):
        y = LunarYear(2023)
        assert y.get_sixty_cycle().get_name() == '癸卯'
        assert y.get_sixty_cycle().get_earth_branch().get_zodiac().get_name() == '兔'

    def test8(self):
        assert LunarYear(1864).get_twenty().get_sixty().get_name() == '上元'

    def test9(self):
        assert LunarYear(1923).get_twenty().get_sixty().get_name() == '上元'

    def test10(self):
        assert LunarYear(1924).get_twenty().get_sixty().get_name() == '中元'

    def test11(self):
        assert LunarYear(1983).get_twenty().get_sixty().get_name() == '中元'

    def test12(self):
        assert LunarYear(1984).get_twenty().get_sixty().get_name() == '下元'

    def test13(self):
        assert LunarYear(2043).get_twenty().get_sixty().get_name() == '下元'

    def test14(self):
        assert LunarYear(1864).get_twenty().get_name() == '一运'

    def test15(self):
        assert LunarYear(1883).get_twenty().get_name() == '一运'

    def test16(self):
        assert LunarYear(1884).get_twenty().get_name() == '二运'

    def test17(self):
        assert LunarYear(1903).get_twenty().get_name() == '二运'

    def test18(self):
        assert LunarYear(1904).get_twenty().get_name() == '三运'

    def test19(self):
        assert LunarYear(1923).get_twenty().get_name() == '三运'

    def test20(self):
        assert LunarYear(2004).get_twenty().get_name() == '八运'

    def test21(self):
        y = LunarYear(1)
        assert y.get_twenty().get_name() == '六运'
        assert y.get_twenty().get_sixty().get_name() == '中元'

    def test22(self):
        y = LunarYear(1863)
        assert y.get_twenty().get_name() == '九运'
        assert y.get_twenty().get_sixty().get_name() == '下元'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarTime


class TestSolarTime(unittest.TestCase):
    def test0(self):
        time: SolarTime = SolarTime(2023, 1, 1, 13, 5, 20)
        assert time.get_name() == '13:05:20'
        assert time.next(-21).get_name() == '13:04:59'
        assert time.next(41).get_name() == '13:06:01'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.culture.phenology import Phenology
from tyme4py.solar import SolarDay, SolarTime


class TestPhenology(unittest.TestCase):
    def test0(self):
        solar_day = SolarDay(2020, 4, 23)
        # 七十二候
        phenology = solar_day.get_phenology_day()
        # 三候
        three_phenology = phenology.get_phenology().get_three_phenology()
        assert solar_day.get_term().get_name() == '谷雨'
        assert three_phenology.get_name() == '初候'
        assert phenology.get_name() == '萍始生'
        # 该候的第5天
        assert phenology.get_day_index() == 4

    def test1(self):
        solar_day = SolarDay(2021, 12, 26)
        # 七十二候
        phenology = solar_day.get_phenology_day()
        # 三候
        three_phenology = phenology.get_phenology().get_three_phenology()
        assert solar_day.get_term().get_name() == '冬至'
        assert three_phenology.get_name() == '二候'
        assert phenology.get_name() == '麋角解'
        # 该候的第1天
        assert phenology.get_day_index() == 0

    def test2(self):
        p = Phenology.from_index(2026, 1)
        jd = p.get_julian_day()
        assert p.get_name() == '麋角解'
        assert jd.get_solar_day().__str__() == '2025年12月26日'
        assert jd.get_solar_time().__str__() == '2025年12月26日 20:49:56'

    def test3(self):
        p = SolarDay.from_ymd(2025, 12, 26).get_phenology()
        jd = p.get_julian_day()
        assert p.get_name() == '麋角解'
        assert jd.get_solar_day().__str__() == '2025年12月26日'
        assert jd.get_solar_time().__str__() == '2025年12月26日 20:49:56'

    def test4(self):
        assert SolarTime.from_ymd_hms(2025, 12, 26, 20, 49, 38).get_phenology().get_name() == '蚯蚓结'
        assert SolarTime.from_ymd_hms(2025, 12, 26, 20, 49, 56).get_phenology().get_name() == '麋角解'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.sixtycycle import HeavenStem


class TestHeavenStem(unittest.TestCase):
    def test(self):
        assert HeavenStem(0).get_name() == '甲'

    def test1(self):
        assert HeavenStem('甲').get_index() == 0

    def test2(self):
        assert HeavenStem('甲').get_element().get_reinforce().get_name() == HeavenStem('丙').get_element().get_name()

    def test3(self):
        assert HeavenStem('甲').get_ten_star(HeavenStem('甲')).get_name() == '比肩'
        assert HeavenStem('甲').get_ten_star(HeavenStem('乙')).get_name() == '劫财'
        assert HeavenStem('甲').get_ten_star(HeavenStem('丙')).get_name() == '食神'
        assert HeavenStem('甲').get_ten_star(HeavenStem('丁')).get_name() == '伤官'
        assert HeavenStem('甲').get_ten_star(HeavenStem('戊')).get_name() == '偏财'
        assert HeavenStem('甲').get_ten_star(HeavenStem('己')).get_name() == '正财'
        assert HeavenStem('甲').get_ten_star(HeavenStem('庚')).get_name() == '七杀'
        assert HeavenStem('甲').get_ten_star(HeavenStem('辛')).get_name() == '正官'
        assert HeavenStem('甲').get_ten_star(HeavenStem('壬')).get_name() == '偏印'
        assert HeavenStem('甲').get_ten_star(HeavenStem('癸')).get_name() == '正印'

    def test4(self):
        assert HeavenStem('庚').get_combine().get_name() == '乙'
        assert HeavenStem('乙').get_combine().get_name() == '庚'
        assert HeavenStem('甲').combine(HeavenStem('己')).get_name() == '土'
        assert HeavenStem('己').combine(HeavenStem('甲')).get_name() == '土'
        assert HeavenStem('丁').combine(HeavenStem('壬')).get_name() == '木'
        assert HeavenStem('壬').combine(HeavenStem('丁')).get_name() == '木'
        assert HeavenStem('甲').combine(HeavenStem('乙')) is None



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import TYPE_CHECKING, Union, List

from tyme4py import LoopTyme, AbstractCulture
from tyme4py.culture import Direction
from tyme4py.enums import Side

if TYPE_CHECKING:
    from tyme4py.lunar import LunarDay, LunarMonth
    from tyme4py.sixtycycle import SixtyCycle, SixtyCycleDay


class FetusEarthBranch(LoopTyme):
    """地支六甲胎神"""
    NAMES: List[str] = ['碓', '厕', '炉', '门', '栖', '床']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> FetusEarthBranch:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> FetusEarthBranch:
        return cls(index)

    def next(self, n: int):
        return FetusEarthBranch(self.next_index(n))


class FetusHeavenStem(LoopTyme):
    """天干六甲胎神"""
    NAMES: List[str] = ['门', '碓磨', '厨灶', '仓库', '房床']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> FetusHeavenStem:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> FetusHeavenStem:
        return cls(index)

    def next(self, n: int) -> FetusHeavenStem:
        return FetusHeavenStem(self.next_index(n))


class FetusMonth(LoopTyme):
    """逐月胎神"""
    NAMES: List[str] = ['占房床', '占户窗', '占门堂', '占厨灶', '占房床', '占床仓', '占碓磨', '占厕户', '占门房', '占房床', '占灶炉', '占房床']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> FetusMonth:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> FetusMonth:
        return cls(index)

    @classmethod
    def from_lunar_month(cls, lunar_month: LunarMonth) -> Union[FetusMonth, None]:
        """
        从农历月初始化
        :param lunar_month: 农历月
        :return: 逐月胎神
        """
        return None if lunar_month.is_leap() else cls(lunar_month.get_month() - 1)

    def next(self, n: int) -> FetusMonth:
        return FetusMonth(self.next_index(n))


class FetusDay(AbstractCulture):
    """逐日胎神"""
    _fetus_heaven_stem: FetusHeavenStem
    _fetus_earth_branch: FetusEarthBranch
    _side: Side
    _direction: Direction

    def __init__(self, sixty_cycle: SixtyCycle):
        self._fetus_heaven_stem = FetusHeavenStem(sixty_cycle.get_heaven_stem().get_index() % 5)
        self._fetus_earth_branch = FetusEarthBranch(sixty_cycle.get_earth_branch().get_index() % 6)
        index: int = [3, 3, 8, 8, 8, 8, 8, 1, 1, 1, 1, 1, 1, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 0, 0, 0, 0, 0, -9, -9, -9, -9, -9, -5, -5, -1, -1, -1, -3, -7, -7, -7, -7, -5, 7, 7, 7, 7, 7, 7, 2, 2, 2, 2, 2, 3, 3, 3, 3][sixty_cycle.get_index()]
        self._side = Side.IN if index < 0 else Side.OUT
        self._direction = Direction(index)

    @classmethod
    def from_lunar_day(cls, lunar_day: LunarDay) -> FetusDay:
        return cls(lunar_day.get_sixty_cycle())

    @classmethod
    def from_sixty_cycle_day(cls, sixty_cycle_day: SixtyCycleDay) -> FetusDay:
        return cls(sixty_cycle_day.get_sixty_cycle())

    def get_name(self) -> str:
        s: str = self._fetus_heaven_stem.get_name() + self._fetus_earth_branch.get_name()
        if '门门' == s:
            s = '占大门'
        elif '碓磨碓' == s:
            s = '占碓磨'
        elif '房床床' == s:
            s = '占房床'
        elif s.startswith('门'):
            s = '占' + s
        s += ' '
        if Side.IN == self._side:
            s += '房内'
        else:
            s += '外'
        direction_name: str = self._direction.get_name()
        if Side.OUT == self._side and '北南西东'.find(direction_name) != -1:
            s += '正'
        s += direction_name
        return s

    def get_side(self) -> Side:
        """
        内外
        :return: 内外
        """
        return self._side

    def get_direction(self) -> Direction:
        """
        方位
        :return: 方位
        """
        return self._direction

    def get_fetus_heaven_stem(self) -> FetusHeavenStem:
        """
        天干六甲胎神
        :return: 天干六甲胎神
        """
        return self._fetus_heaven_stem

    def get_fetus_earth_branch(self) -> FetusEarthBranch:
        """
        地支六甲胎神
        :return: 地支六甲胎神
        """
        return self._fetus_earth_branch



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import List, Union

from tyme4py import LoopTyme
from tyme4py.culture import Luck, Element


class MinorRen(LoopTyme):
    """小六壬"""
    NAMES: List[str] = ['大安', '留连', '速喜', '赤口', '小吉', '空亡']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    def next(self, n: int) -> MinorRen:
        return MinorRen(self.next_index(n))

    def get_luck(self) -> Luck:
        """
        吉凶
        :return: 吉凶
        """
        return Luck(self.get_index() % 2)

    def get_element(self) -> Element:
        """
        五行
        :return: 五行
        """
        return Element([0, 4, 1, 3, 0, 2][self.get_index()])



# -*- coding:utf-8 -*-
import unittest

from tyme4py.sixtycycle import ThreePillars
from tyme4py.solar import SolarDay


class TestThreePillars(unittest.TestCase):
    def test(self):
        t: ThreePillars = ThreePillars('甲戌', '甲戌', '甲戌')
        day_list: [str] = []
        for d in t.get_solar_days(1, 2200):
            day_list.append(d.__str__())
        assert day_list == [
            '14年10月17日',
            '194年11月1日',
            '254年10月17日',
            '434年11月1日',
            '494年10月17日',
            '674年11月1日',
            '734年10月17日',
            '794年10月2日',
            '974年10月17日',
            '1034年10月2日',
            '1214年10月17日',
            '1274年10月2日',
            '1454年10月17日',
            '1514年10月2日',
            '1694年10月27日',
            '1754年10月13日',
            '1934年10月30日',
            '1994年10月15日',
            '2174年10月31日'
        ]

    def test1(self):
        assert SolarDay.from_ymd(1034, 10, 2).get_sixty_cycle_day().get_three_pillars().get_name() == '甲戌 甲戌 甲戌'



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import TYPE_CHECKING, List, Union

from tyme4py import LoopTyme, AbstractCulture

if TYPE_CHECKING:
    from tyme4py.sixtycycle import SixtyCycle


class PengZuEarthBranch(LoopTyme):
    """地支彭祖百忌"""
    NAMES: List[str] = ['子不问卜自惹祸殃', '丑不冠带主不还乡', '寅不祭祀神鬼不尝', '卯不穿井水泉不香', '辰不哭泣必主重丧', '巳不远行财物伏藏', '午不苫盖屋主更张', '未不服药毒气入肠', '申不安床鬼祟入房', '酉不会客醉坐颠狂', '戌不吃犬作怪上床', '亥不嫁娶不利新郎']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> PengZuEarthBranch:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> PengZuEarthBranch:
        return cls(index)

    def next(self, n: int):
        return PengZuEarthBranch(self.next_index(n))


class PengZuHeavenStem(LoopTyme):
    """天干彭祖百忌"""
    NAMES: List[str] = ['甲不开仓财物耗散', '乙不栽植千株不长', '丙不修灶必见灾殃', '丁不剃头头必生疮', '戊不受田田主不祥', '己不破券二比并亡', '庚不经络织机虚张', '辛不合酱主人不尝', '壬不泱水更难提防', '癸不词讼理弱敌强']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> PengZuHeavenStem:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> PengZuHeavenStem:
        return cls(index)

    def next(self, n: int) -> PengZuHeavenStem:
        return PengZuHeavenStem(self.next_index(n))


class PengZu(AbstractCulture):
    """彭祖百忌"""
    _peng_zu_heaven_stem: PengZuHeavenStem
    """天干彭祖百忌"""
    _peng_zu_earth_branch: PengZuEarthBranch
    """地支彭祖百忌"""

    def __init__(self, sixty_cycle: SixtyCycle):
        self._peng_zu_heaven_stem = PengZuHeavenStem(sixty_cycle.get_heaven_stem().get_index())
        self._peng_zu_earth_branch = PengZuEarthBranch(sixty_cycle.get_earth_branch().get_index())

    @classmethod
    def from_sixty_cycle(cls, sixty_cycle: SixtyCycle) -> PengZu:
        return cls(sixty_cycle)

    def get_name(self) -> str:
        return self._peng_zu_heaven_stem.get_name() + " " + self._peng_zu_earth_branch.get_name()

    def get_peng_zu_heaven_stem(self) -> PengZuHeavenStem:
        """
        天干彭祖百忌
        :return: 天干彭祖百忌
        """
        return self._peng_zu_heaven_stem

    def get_peng_zu_earth_branch(self) -> PengZuEarthBranch:
        """
        地支彭祖百忌
        :return: 地支彭祖百忌
        """
        return self._peng_zu_earth_branch



# -*- coding:utf-8 -*-
import unittest

from tyme4py.culture import Zodiac
from tyme4py.rabbyung import RabByungYear, RabByungElement


class TestRabByungYear(unittest.TestCase):
    def test0(self):
        y: RabByungYear = RabByungYear.from_element_zodiac(0, RabByungElement.from_name("火"), Zodiac.from_name("兔"))
        assert "第一饶迥火兔年" == y.get_name()
        assert "1027年" == y.get_solar_year().get_name()
        assert "丁卯" == y.get_sixty_cycle().get_name()
        assert 10 == y.get_leap_month()

    def test1(self):
        assert "第一饶迥火兔年" == RabByungYear.from_year(1027).get_name()

    def test2(self):
        assert "第十七饶迥铁虎年" == RabByungYear.from_year(2010).get_name()

    def test3(self):
        assert 5 == RabByungYear.from_year(2043).get_leap_month()
        assert 0 == RabByungYear.from_year(2044).get_leap_month()

    def test4(self):
        assert "第十六饶迥铁牛年" == RabByungYear.from_year(1961).get_name()



# -*- coding:utf-8 -*-
from __future__ import annotations

import re
from typing import TYPE_CHECKING, List, Union

from tyme4py import AbstractTyme
from tyme4py.enums import FestivalType

if TYPE_CHECKING:
    from tyme4py.lunar import LunarDay
    from tyme4py.solar import SolarTerm, SolarDay


class LunarFestival(AbstractTyme):
    """
    农历传统节日--依据国家标准《农历的编算和颁行》GB/T 33661-2017
    农历传统节日有：春节、元宵节、龙头节、上巳节、清明节、端午节、七夕节、中元节、中秋节、重阳节、冬至节、腊八节、除夕。
    """
    NAMES: List[str] = ['春节', '元宵节', '龙头节', '上巳节', '清明节', '端午节', '七夕节', '中元节', '中秋节', '重阳节', '冬至节', '腊八节', '除夕']
    """名称"""
    DATA: str = '@0000101@0100115@0200202@0300303@04107@0500505@0600707@0700715@0800815@0900909@10124@1101208@122'
    _type: FestivalType
    """类型"""
    _index: int
    """索引"""
    _day: LunarDay
    """农历日"""
    _name: str
    """名称"""
    _solar_term: Union[SolarTerm, None]
    """节气"""

    def __init__(self, festival_type: FestivalType, day: LunarDay, solar_term: Union[SolarTerm, None], data: Union[str, None] = None):
        """
        通过农历日 LunarDay得到
        :param festival_type: 节日类型
        :param day: 农历日
        :param solar_term: 节气
        :param data:
        """
        self._type = festival_type
        self._day = day
        self._solar_term = solar_term
        self._index = int(data[1: 3], 10)
        self._name = LunarFestival.NAMES[self._index]

    @classmethod
    def from_index(cls, year: int, index: int) -> Union[LunarFestival, None]:
        """
        指定索引得到
        :param year: 年
        :param index: 索引
        :return:
        """
        if index < 0 or index >= cls.NAMES.__len__():
            return None
        from tyme4py.lunar import LunarDay
        from tyme4py.solar import SolarTerm
        pattern = re.compile(f"@{index:02}\\d+")
        matcher = pattern.search(cls.DATA)
        if matcher:
            data: str = matcher.group()
            festival_type: int = ord(data[3]) - 48
            if festival_type == 0:
                return cls(FestivalType.DAY, LunarDay(year, int(data[4: 6], 10), int(data[6:], 10)), None, data)
            elif festival_type == 1:
                solar_term: SolarTerm = SolarTerm(year, int(data[4:], 10))
                return cls(FestivalType.TERM, solar_term.get_solar_day().get_lunar_day(), solar_term, data)
            elif festival_type == 2:
                return cls(FestivalType.EVE, LunarDay(year + 1, 1, 1).next(-1), None, data)
            else:
                return None
        return None

    @classmethod
    def from_ymd(cls, year: int, month: int, day: int) -> Union[LunarFestival, None]:
        """
        指定农历年、月、日得到
        :param year: 农历年
        :param month: 农历月
        :param day: 农历日
        :return:
        """
        from tyme4py.lunar import LunarDay
        from tyme4py.solar import SolarTerm
        pattern = re.compile("@\\d{2}0" + "{:02}{:02}".format(month, day))
        matcher = pattern.search(cls.DATA)
        if matcher:
            return cls(FestivalType.DAY, LunarDay(year, month, day), None, matcher.group())
        reg = re.compile('@\\d{2}1\\d{2}')
        matcher = reg.search(cls.DATA)
        while matcher:
            data: str = matcher.group()
            solar_term: SolarTerm = SolarTerm(year, int(data[4:], 10))
            lunar_day: LunarDay = solar_term.get_solar_day().get_lunar_day()
            if lunar_day.get_year() == year and lunar_day.get_month() == month and lunar_day.get_day() == day:
                return cls(FestivalType.TERM, lunar_day, solar_term, data)
            last_position = matcher.end()
            matcher = reg.match(cls.DATA, last_position)
        pattern = re.compile("@\\d{2}2")
        matcher = pattern.search(cls.DATA)
        if matcher:
            lunar_day: LunarDay = LunarDay(year, month, day)
            next_day: LunarDay = lunar_day.next(1)
            if next_day.get_month() == 1 and next_day.get_day() == 1:
                return cls(FestivalType.EVE, lunar_day, None, matcher.group())
        return None

    def get_name(self) -> str:
        return self._name

    def get_index(self) -> int:
        """
        :return: 索引
        """
        return self._index

    def get_day(self) -> LunarDay:
        """
        :return: 农历日
        """
        return self._day

    def get_type(self) -> FestivalType:
        """
        :return: 节日类型
        """
        return self._type

    def get_solar_term(self) -> Union[SolarTerm, None]:
        """
        节气，非节气返回None
        :return:  节气
        """
        return self._solar_term

    def __str__(self) -> str:
        return f'{self._day} {self._name}'

    def next(self, n: int) -> LunarFestival:
        size: int = LunarFestival.NAMES.__len__()
        i: int = self._index + n
        return LunarFestival.from_index(int((self._day.get_year() * size + i) / size), self.index_of(i, size))


class SolarFestival(AbstractTyme):
    """公历现代节日有：元旦、三八妇女节、植树节、五一劳动节、五四青年节、六一儿童节、建党节、八一建军节、教师节、国庆节。"""
    NAMES: List[str] = ['元旦', '三八妇女节', '植树节', '五一劳动节', '五四青年节', '六一儿童节', '建党节', '八一建军节', '教师节', '国庆节']
    DATA: str = '@00001011950@01003081950@02003121979@03005011950@04005041950@05006011950@06007011941@07008011933@08009101985@09010011950'
    _type: FestivalType
    """类型"""
    _index: int
    """索引"""
    _day: SolarDay
    """公历日"""
    _name: str
    """名称"""
    _start_year: int
    """起始年"""

    def __init__(self, festival_type: FestivalType, day: SolarDay, start_year: int, data: str):
        """
        通过公历日 SolarDay得到
        :param festival_type:   节日类型
        :param day: 农历日
        :param start_year: 开始年份
        :param data:
        """
        self._type = festival_type
        self._day = day
        self._start_year = start_year
        self._index = int(data[1: 3], 10)
        self._name = self.NAMES[self._index]

    @classmethod
    def from_index(cls, year: int, index: int) -> Union[SolarFestival, None]:
        """
        指定索引得到
        :param year:  年份
        :param index: 索引
        :return:
        """
        if index < 0 or index >= cls.NAMES.__len__():
            return None
        from tyme4py.solar import SolarDay
        pattern = re.compile(f"@{index:02}\\d+")
        matcher = pattern.search(cls.DATA)
        if matcher:
            data: str = matcher.group()
            type: int = ord(data[3]) - 48
            if type == 0:
                start_year: int = int(data[8:], 10)
                if year >= start_year:
                    return cls(FestivalType.DAY, SolarDay(year, int(data[4: 6], 10), int(data[6: 8], 10)), start_year, data)
        return None

    @classmethod
    def from_ymd(cls, year: int, month: int, day: int) -> Union[SolarFestival, None]:
        """
        指定年、月、日得到
        :param year:  年
        :param month: 月
        :param day:   日
        :return:
        """
        from tyme4py.solar import SolarDay
        pattern = re.compile("@\\d{2}0" + f"{month:02}{day:02}\\d+")
        matcher = pattern.search(cls.DATA)
        if matcher:
            data: str = matcher.group()
            start_year: int = int(data[8:], 10)
            if year >= start_year:
                return cls(FestivalType.DAY, SolarDay(year, month, day), start_year, data)
        return None

    def get_name(self) -> str:
        return self._name

    def get_index(self) -> int:
        """
        :return: 索引
        """
        return self._index

    def get_day(self) -> SolarDay:
        """
        :return: 公历日
        """
        return self._day

    def get_type(self) -> FestivalType:
        """
        :return: 节日类型
        """
        return self._type

    def get_start_year(self) -> int:
        """
        起始年
        :return: 年
        """
        return self._start_year

    def __str__(self) -> str:
        return f'{self._day} {self._name}'

    def next(self, n: int) -> Union[SolarFestival, None]:
        size: int = SolarFestival.NAMES.__len__()
        i: int = self._index + n
        return SolarFestival.from_index(int((self._day.get_year() * size + i) / size), self.index_of(i, size))



# -*- coding:utf-8 -*-
from enum import Enum


class YinYang(Enum):
    """阴阳：YIN=0=阴，YANG=1=阳"""
    YIN = 0
    """阴"""
    YANG = 1
    """阳"""


class Side(Enum):
    """内外：IN=0=内，OUT=1=外"""
    IN = 0
    """内"""
    OUT = 1
    """外"""


class Gender(Enum):
    """性别：WOMAN=0=女，MAN=1=男"""
    WOMAN = 0
    """女"""
    MAN = 1
    """男"""


class FestivalType(Enum):
    """节日类型：DAY=0=日期，TERM=1=节气，EVE=2=除夕"""
    DAY = 0
    """日期"""
    TERM = 1
    """节气"""
    EVE = 2
    """除夕"""


class HideHeavenStemType(Enum):
    """藏干类型：RESIDUAL=0=余气，MIDDLE=1=中气，MAIN=2=本气"""
    RESIDUAL = 0
    """余气"""
    MIDDLE = 1
    """中气"""
    MAIN = 2
    """本气"""



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestDuty(unittest.TestCase):
    def test(self):
        assert SolarDay(2023, 10, 30).get_lunar_day().get_duty().get_name() == '闭'

    def test1(self):
        assert SolarDay(2023, 10, 19).get_lunar_day().get_duty().get_name() == '建'

    def test2(self):
        assert SolarDay(2023, 10, 7).get_lunar_day().get_duty().get_name() == '除'

    def test3(self):
        assert SolarDay(2023, 10, 8).get_lunar_day().get_duty().get_name() == '除'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay, SolarTime


class TestTaboo(unittest.TestCase):
    def test0(self):
        taboos: [str] = []
        recommends = SolarDay(2024, 6, 26).get_lunar_day().get_recommends()
        for t in recommends:
            taboos.append(t.get_name())
        assert taboos == ['嫁娶', '祭祀', '理发', '作灶', '修饰垣墙', '平治道涂', '整手足甲', '沐浴', '冠笄']

    def test1(self):
        taboos: [str] = []
        avoids = SolarDay(2024, 6, 26).get_lunar_day().get_avoids()
        for t in avoids:
            taboos.append(t.get_name())
        assert taboos, ['破土', '出行', '栽种']

    def test2(self):
        taboos: [str] = []
        recommends = SolarTime(2024, 6, 25, 4, 0, 0).get_lunar_hour().get_recommends()
        for t in recommends:
            taboos.append(t.get_name())
        assert taboos == []

    def test3(self):
        taboos: [str] = []
        avoids = SolarTime(2024, 6, 25, 4, 0, 0).get_lunar_hour().get_avoids()
        for t in avoids:
            taboos.append(t.get_name())
        assert taboos == ['诸事不宜']

    def test4(self):
        taboos: [str] = []
        recommends = SolarTime(2024, 4, 22, 0, 0, 0).get_lunar_hour().get_recommends()
        for t in recommends:
            taboos.append(t.get_name())
        assert taboos == ['嫁娶', '交易', '开市', '安床', '祭祀', '求财']

    def test5(self):
        taboos: [str] = []
        avoids = SolarTime(2024, 4, 22, 0, 0, 0).get_lunar_hour().get_avoids()
        for t in avoids:
            taboos.append(t.get_name())
        assert taboos, ['出行', '移徙', '赴任', '词讼', '祈福', '修造', '求嗣']

    def test6(self):
        taboos: [str] = []
        recommends = SolarDay(2021, 3, 7).get_lunar_day().get_recommends()
        for t in recommends:
            taboos.append(t.get_name())
        assert taboos == ['裁衣', '经络', '伐木', '开柱眼', '拆卸', '修造', '动土', '上梁', '合脊', '合寿木', '入殓', '除服', '成服', '移柩', '破土', '安葬', '启钻', '修坟', '立碑']



# -*- coding:utf-8 -*-
import unittest

from tyme4py.sixtycycle import SixtyCycle, EarthBranch, HeavenStem


class TestSixtyCycle(unittest.TestCase):
    def test0(self):
        assert SixtyCycle(13).get_name() == '丁丑'

    def test1(self):
        assert SixtyCycle('丁丑').get_index() == 13

    def test2(self):
        """五行"""
        assert SixtyCycle('辛酉').get_sound().get_name() == '石榴木'
        assert SixtyCycle('癸酉').get_sound().get_name() == '剑锋金'
        assert SixtyCycle('己亥').get_sound().get_name() == '平地木'

    def test3(self):
        """旬"""
        assert SixtyCycle('甲子').get_ten().get_name() == '甲子'
        assert SixtyCycle('乙卯').get_ten().get_name() == '甲寅'
        assert SixtyCycle('癸巳').get_ten().get_name() == '甲申'

    def test4(self):
        """旬空"""
        assert SixtyCycle('甲子').get_extra_earth_branches() == [EarthBranch('戌'), EarthBranch('亥')]
        assert SixtyCycle('乙卯').get_extra_earth_branches() == [EarthBranch('子'), EarthBranch('丑')]
        assert SixtyCycle('癸巳').get_extra_earth_branches() == [EarthBranch('午'), EarthBranch('未')]

    def test5(self):
        """地势(长生十二神)"""
        assert HeavenStem('丙').get_terrain(EarthBranch('寅')).get_name() == '长生'
        assert HeavenStem('辛').get_terrain(EarthBranch('亥')).get_name() == '沐浴'



# -*- coding:utf-8 -*-
from __future__ import annotations

import warnings
from math import ceil
from typing import TYPE_CHECKING, Union, List

from tyme4py import AbstractCulture, AbstractTyme
from tyme4py.culture import Duty
from tyme4py.eightchar.provider import ChildLimitProvider
from tyme4py.eightchar.provider.impl import DefaultChildLimitProvider
from tyme4py.enums import Gender, YinYang

if TYPE_CHECKING:
    from tyme4py.lunar import LunarYear
    from tyme4py.solar import SolarTime, SolarTerm, SolarDay
    from tyme4py.sixtycycle import SixtyCycle, SixtyCycleYear, ThreePillars


class EightChar(AbstractCulture):
    """
    八字
    """
    _three_pillars: ThreePillars
    """三柱"""
    _hour: SixtyCycle
    """时柱"""

    def __init__(self, year: Union[SixtyCycle, str], month: Union[SixtyCycle, str], day: Union[SixtyCycle, str], hour: Union[SixtyCycle, str]):
        """
        :param year: 年干支
        :param month: 月干支
        :param day: 日干支
        :param hour: 时干支
        """
        from tyme4py.sixtycycle import SixtyCycle, ThreePillars
        self._three_pillars = ThreePillars(year if isinstance(year, SixtyCycle) else SixtyCycle(year), month if isinstance(month, SixtyCycle) else SixtyCycle(month), day if isinstance(day, SixtyCycle) else SixtyCycle(day))
        self._hour = hour if isinstance(hour, SixtyCycle) else SixtyCycle(hour)

    def get_year(self) -> SixtyCycle:
        """
        年柱
        :return: 干支 SixtyCycle
        """
        return self._three_pillars.get_year()

    def get_month(self) -> SixtyCycle:
        """
        月柱
        :return: 干支 SixtyCycle
        """
        return self._three_pillars.get_month()

    def get_day(self) -> SixtyCycle:
        """
        日柱
        :return:  干支 SixtyCycle
        """
        return self._three_pillars.get_day()

    def get_hour(self) -> SixtyCycle:
        """
        时柱
        :return: 干支 SixtyCycle
        """
        return self._hour

    def get_fetal_origin(self) -> SixtyCycle:
        """
        胎元
        :return: 干支 SixtyCycle
        """
        from tyme4py.sixtycycle import SixtyCycle
        m = self.get_month()
        return SixtyCycle(m.get_heaven_stem().next(1).get_name() + m.get_earth_branch().next(3).get_name())

    def get_fetal_breath(self) -> SixtyCycle:
        """
        胎息
        :return: 干支 SixtyCycle
        """
        from tyme4py.sixtycycle import SixtyCycle, EarthBranch
        d = self.get_day()
        return SixtyCycle(d.get_heaven_stem().next(5).get_name() + EarthBranch(13 - d.get_earth_branch().get_index()).get_name())

    def get_own_sign(self) -> SixtyCycle:
        """
        命宫
        :return: 干支 SixtyCycle
        """
        from tyme4py.sixtycycle import SixtyCycle, EarthBranch, HeavenStem
        m: int = self.get_month().get_earth_branch().get_index() - 1
        if m < 1:
            m += 12
        h: int = self._hour.get_earth_branch().get_index() - 1
        if h < 1:
            h += 12
        offset: int = m + h
        offset = (26 if offset >= 14 else 14) - offset
        return SixtyCycle(HeavenStem((self.get_year().get_heaven_stem().get_index() + 1) * 2 + offset - 1).get_name() + EarthBranch(offset + 1).get_name())

    def get_body_sign(self) -> SixtyCycle:
        """
        身宫
        :return: 干支 SixtyCycle
        """
        from tyme4py.sixtycycle import SixtyCycle, EarthBranch, HeavenStem
        offset: int = self.get_month().get_earth_branch().get_index() - 1
        if offset < 1:
            offset += 12
        offset += self._hour.get_earth_branch().get_index() + 1
        if offset > 12:
            offset -= 12
        return SixtyCycle(HeavenStem((self.get_year().get_heaven_stem().get_index() + 1) * 2 + offset - 1).get_name() + EarthBranch(offset + 1).get_name())

    def get_duty(self) -> Duty:
        """
        建除十二值神
        :return: 建除十二值神
        """
        warnings.warn('get_duty() is deprecated, please use SixtyCycleDay.get_duty() instead.', DeprecationWarning)
        return Duty(self.get_day().get_earth_branch().get_index() - self.get_month().get_earth_branch().get_index())

    def get_name(self) -> str:
        return f'{self._three_pillars} {self._hour}'

    def get_solar_times(self, start_year: int, end_year: int) -> List[SolarTime]:
        """
        八字转公历时刻列表
        :param start_year: 开始年份，支持1-9999年
        :param end_year: 结束年份，支持1-9999年
        :return: 公历时刻 SolarTime的列表
        """
        from tyme4py.sixtycycle import HeavenStem
        from tyme4py.solar import SolarTime, SolarTerm
        l: List[SolarTime] = []
        year = self.get_year()
        month = self.get_month()
        day = self.get_day()
        # 月地支距寅月的偏移值
        m: int = month.get_earth_branch().next(-2).get_index()
        # 月天干要一致
        if not HeavenStem((year.get_heaven_stem().get_index() + 1) * 2 + m) == month.get_heaven_stem():
            return l

        # 1年的立春是辛酉，序号57
        y: int = year.next(-57).get_index() + 1
        # 节令偏移值
        m *= 2
        # 时辰地支转时刻
        h: int = self._hour.get_earth_branch().get_index() * 2
        # 兼容子时多流派
        hours: List[int] = [0, 23] if h == 0 else [h]
        base_year: int = start_year - 1
        if base_year > y:
            y += 60 * int(ceil((base_year - y) / 60.0))

        while y <= end_year:
            # 立春为寅月的开始
            term: SolarTerm = SolarTerm(y, 3)
            # 节令推移，年干支和月干支就都匹配上了
            if m > 0:
                term = term.next(m)

            solar_time: SolarTime = term.get_julian_day().get_solar_time()
            if solar_time.get_year() >= start_year:
                # 日干支和节令干支的偏移值
                solar_day: SolarDay = solar_time.get_solar_day()
                d: int = day.next(-solar_day.get_lunar_day().get_sixty_cycle().get_index()).get_index()
                if d > 0:
                    # 从节令推移天数
                    solar_day = solar_day.next(d)
                for i in range(0, hours.__len__()):
                    mi: int = 0
                    s: int = 0
                    hour: int = hours[i]
                    if d == 0 and hour == solar_time.get_hour():
                        # 如果正好是节令当天，且小时和节令的小时数相等的极端情况，把分钟和秒钟带上
                        mi = solar_time.get_minute()
                        s = solar_time.get_second()

                    time: SolarTime = SolarTime(solar_day.get_year(), solar_day.get_month(), solar_day.get_day(), hour, mi, s)
                    if d == 30:
                        time = time.next(-3600)
                    # 验证一下
                    if time.get_lunar_hour().get_eight_char() == self:
                        l.append(time)
            y += 60
        return l


class ChildLimitInfo:
    """童限信息"""
    _start_time: SolarTime
    """开始(即出生)的公历时刻"""
    _end_time: SolarTime
    """结束(即开始起运)的公历时刻"""
    _year_count: int
    """年数"""
    _month_count: int
    """月数"""
    _day_count: int
    """日数"""
    _hour_count: int
    """小时数"""
    _minute_count: int
    """分钟数"""

    def __init__(self, start_time: SolarTime, end_time: SolarTime, year_count: int, month_count: int, day_count: int, hour_count: int, minute_count: int):
        self._start_time = start_time
        self._end_time = end_time
        self._year_count = year_count
        self._month_count = month_count
        self._day_count = day_count
        self._hour_count = hour_count
        self._minute_count = minute_count

    def get_start_time(self) -> SolarTime:
        """
        :return:  开始(即出生)的公历时刻
        """
        return self._start_time

    def get_end_time(self) -> SolarTime:
        """
        :return: 结束(即开始起运)的公历时刻
        """
        return self._end_time

    def get_year_count(self) -> int:
        """
        :return: 年数
        """
        return self._year_count

    def get_month_count(self) -> int:
        """
        :return: 月数
        """
        return self._month_count

    def get_day_count(self) -> int:
        """
        :return:  日数
        """
        return self._day_count

    def get_hour_count(self) -> int:
        """
        :return: 小时数
        """
        return self._hour_count

    def get_minute_count(self) -> int:
        """
        :return: 分钟数
        """
        return self._minute_count


class ChildLimit:
    """童限（从出生到起运的时间段）"""
    provider: ChildLimitProvider = DefaultChildLimitProvider()
    """童限计算接口"""
    _eight_char: EightChar
    """八字"""
    _gender: Gender
    """性别"""
    _forward: bool
    """顺逆"""
    _info: ChildLimitInfo
    """童限信息"""

    def __init__(self, birth_time: SolarTime, gender: Gender):
        """
        通过出生公历时刻初始化
        :param birth_time: 出生公历时刻
        :param gender: 性别
        """
        self._gender = gender
        self._eight_char = birth_time.get_lunar_hour().get_eight_char()
        # 阳男阴女顺推，阴男阳女逆推
        yang: bool = YinYang.YANG == self._eight_char.get_year().get_heaven_stem().get_yin_yang()
        man: bool = Gender.MAN == gender
        self._forward = (yang and man) or (not yang and not man)
        term: SolarTerm = birth_time.get_term()
        if not term.is_jie():
            term = term.next(-1)
        if self._forward:
            term = term.next(2)
        self._info = ChildLimit.provider.get_info(birth_time, term)

    def get_eight_char(self) -> EightChar:
        """
        :return: 八字
        """
        return self._eight_char

    def get_gender(self) -> Gender:
        """
        :return: 性别
        """
        return self._gender

    def get_year_count(self) -> int:
        """
        :return: 年数
        """
        return self._info.get_year_count()

    def get_month_count(self) -> int:
        """
        :return: 月数
        """
        return self._info.get_month_count()

    def get_day_count(self) -> int:
        """
        :return: 日数
        """
        return self._info.get_day_count()

    def get_hour_count(self) -> int:
        """
        :return: 小时数
        """
        return self._info.get_hour_count()

    def get_minute_count(self) -> int:
        """
        :return: 分钟数
        """
        return self._info.get_minute_count()

    def get_start_time(self) -> SolarTime:
        """
        开始(即出生)的公历时刻
        :return: 公历时刻
        """
        return self._info.get_start_time()

    def get_end_time(self) -> SolarTime:
        """
        结束(即开始起运)的公历时刻
        :return: 公历时刻
        """
        return self._info.get_end_time()

    def is_forward(self) -> bool:
        """
        是否顺推
        :return: true/false
        """
        return self._forward

    def get_start_decade_fortune(self) -> DecadeFortune:
        """
        :return: 起运大运
        """
        return DecadeFortune(self, 0)

    def get_decade_fortune(self) -> DecadeFortune:
        """
        :return: 所属大运
        """
        return DecadeFortune(self, -1)

    def get_start_fortune(self) -> Fortune:
        """
        :return: 小运
        """
        return Fortune(self, 0)

    def get_end_lunar_year(self) -> LunarYear:
        """
        结束农历年
        :return: 农历年
        """
        warnings.warn('get_end_lunar_year() is deprecated, please use SixtyCycleDay.get_duty() instead.', DeprecationWarning)
        from tyme4py.lunar import LunarYear
        return LunarYear.from_year(self.get_start_time().get_lunar_hour().get_year() + self.get_end_time().get_year() - self.get_start_time().get_year())

    def get_start_sixty_cycle_year(self) -> SixtyCycleYear:
        """
        开始(即出生)干支年
        :return: 干支年
        """
        from tyme4py.sixtycycle import SixtyCycleYear
        return SixtyCycleYear.from_year(self.get_start_time().get_year())

    def get_end_sixty_cycle_year(self) -> SixtyCycleYear:
        """
        结束(即起运)干支年
        :return: 干支年
        """
        from tyme4py.sixtycycle import SixtyCycleYear
        return SixtyCycleYear.from_year(self.get_end_time().get_year())

    def get_start_age(self) -> int:
        """
        开始年龄
        :return: 数字
        """
        return 1

    def get_end_age(self) -> int:
        """
        结束年龄
        :return: 数字
        """
        n: int = self.get_end_sixty_cycle_year().get_year() - self.get_start_sixty_cycle_year().get_year()
        return max(n, 1)


class Fortune(AbstractTyme):
    """
    小运--在十年大运中，每一年为一小运。童限结束的公历时刻，既是大运的开始，也是小运的开始。
    """
    _child_limit: ChildLimit
    """童限"""
    _index: int
    """序号"""

    def __init__(self, child_limit: ChildLimit, index: int):
        """
        通过童限初始化
        :param child_limit: 童限
        :param index: 序号
        """
        self._child_limit = child_limit
        self._index = index

    def get_age(self) -> int:
        """
        :return: 年龄
        """
        return self._child_limit.get_end_sixty_cycle_year().get_year() - self._child_limit.get_start_sixty_cycle_year().get_year() + 1 + self._index

    def get_lunar_year(self) -> LunarYear:
        """
        :return: 农历年
        """
        warnings.warn('get_lunar_year() is deprecated, please use get_sixty_cycle_year() instead.', DeprecationWarning)
        return self._child_limit.get_end_lunar_year().next(self._index)

    def get_sixty_cycle_year(self) -> SixtyCycleYear:
        """
        :return: 干支年
        """
        return self._child_limit.get_end_sixty_cycle_year().next(self._index)

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        :return: 干支
        """
        n: int = self.get_age()
        return self._child_limit.get_eight_char().get_hour().next(n if self._child_limit.is_forward() else -n)

    def get_name(self) -> str:
        return self.get_sixty_cycle().get_name()

    def next(self, n: int) -> Fortune:
        return Fortune(self._child_limit, self._index + n)


class DecadeFortune(AbstractTyme):
    """
    大运（10年1大运）--自起运开始，每十年为一大运。童限结束的公历时刻，即开始起运，是大运的开始。
    """
    _child_limit: ChildLimit  # 童限
    _index: int  # 序号

    def __init__(self, child_limit: ChildLimit, index: int):
        """
        通过童限初始化
        :param child_limit: 童限
        :param index: 序号
        """
        self._child_limit = child_limit
        self._index = index

    def get_start_age(self) -> int:
        """
        :return: 开始年龄
        """
        return self._child_limit.get_end_sixty_cycle_year().get_year() - self._child_limit.get_start_sixty_cycle_year().get_year() + 1 + self._index * 10

    def get_end_age(self) -> int:
        """
        :return: 结束年龄
        """
        return self.get_start_age() + 9

    def get_start_lunar_year(self) -> LunarYear:
        """
        开始农历年
        :return: 农历年
        """
        warnings.warn('get_lunar_year() is deprecated, please use get_sixty_cycle_year() instead.', DeprecationWarning)
        return self._child_limit.get_end_lunar_year().next(self._index * 10)

    def get_end_lunar_year(self) -> LunarYear:
        """
        结束农历年
        :return: 农历年
        """
        warnings.warn('get_lunar_year() is deprecated, please use get_sixty_cycle_year() instead.', DeprecationWarning)
        return self.get_start_lunar_year().next(9)

    def get_start_sixty_cycle_year(self) -> SixtyCycleYear:
        """
        开始干支年
        :return: 干支年
        """
        return self._child_limit.get_end_sixty_cycle_year().next(self._index * 10)

    def get_end_sixty_cycle_year(self) -> SixtyCycleYear:
        """
        结束干支年
        :return: 干支年
        """
        return self.get_start_sixty_cycle_year().next(9)

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        :return: 干支
        """
        return self._child_limit.get_eight_char().get_month().next(self._index + 1 if self._child_limit.is_forward() else -self._index - 1)

    def get_name(self) -> str:
        return self.get_sixty_cycle().get_name()

    def next(self, n: int) -> DecadeFortune:
        return DecadeFortune(self._child_limit, self._index + n)

    def get_start_fortune(self) -> Fortune:
        """
        本轮大运中开始的小运
        :return: 小运 Fortune
        """
        return Fortune(self._child_limit, self._index * 10)



# -*- coding:utf-8 -*-
import unittest

from tyme4py.lunar import LunarDay


class TestLunarDay(unittest.TestCase):
    def test1(self):
        assert LunarDay(0, 11, 18).get_solar_day().__str__() == '1年1月1日'

    def test2(self):
        assert LunarDay(9999, 12, 2).get_solar_day().__str__() == '9999年12月31日'

    def test3(self):
        assert LunarDay(1905, 1, 1).get_solar_day().__str__() == '1905年2月4日'

    def test4(self):
        assert LunarDay(2038, 12, 29).get_solar_day().__str__() == '2039年1月23日'

    def test5(self):
        assert LunarDay(1500, 1, 1).get_solar_day().__str__() == '1500年1月31日'

    def test6(self):
        assert LunarDay(1500, 12, 29).get_solar_day().__str__() == '1501年1月18日'

    def test7(self):
        assert LunarDay(1582, 9, 18).get_solar_day().__str__() == '1582年10月4日'

    def test8(self):
        assert LunarDay(1582, 9, 19).get_solar_day().__str__() == '1582年10月15日'

    def test9(self):
        assert LunarDay(2019, 12, 12).get_solar_day().__str__() == '2020年1月6日'

    def test10(self):
        assert LunarDay(2033, -11, 1).get_solar_day().__str__() == '2033年12月22日'

    def test11(self):
        assert LunarDay(2021, 6, 7).get_solar_day().__str__() == '2021年7月16日'

    def test12(self):
        assert LunarDay(2034, 1, 1).get_solar_day().__str__() == '2034年2月19日'

    def test13(self):
        assert LunarDay(2033, 12, 1).get_solar_day().__str__() == '2034年1月20日'

    def test14(self):
        assert LunarDay(7013, -11, 4).get_solar_day().__str__() == '7013年12月24日'

    def test15(self):
        assert LunarDay(2023, 8, 24).get_sixty_cycle().__str__() == '己亥'

    def test16(self):
        assert LunarDay(1653, 1, 6).get_sixty_cycle().__str__() == '癸酉'

    def test17(self):
        assert LunarDay(2010, 1, 1).next(31).__str__() == '农历庚寅年二月初二'

    def test18(self):
        assert LunarDay(2012, 3, 1).next(60).__str__() == '农历壬辰年闰四月初一'

    def test19(self):
        assert LunarDay(2012, 3, 1).next(88).__str__() == '农历壬辰年闰四月廿九'

    def test20(self):
        assert LunarDay(2012, 3, 1).next(89).__str__() == '农历壬辰年五月初一'

    def test21(self):
        assert LunarDay(2020, 4, 1).get_solar_day().__str__() == '2020年4月23日'

    def test22(self):
        assert LunarDay(2024, 1, 1).get_lunar_month().get_lunar_year().get_sixty_cycle().get_name() == '甲辰'

    def test23(self):
        assert LunarDay(2023, 12, 30).get_lunar_month().get_lunar_year().get_sixty_cycle().get_name() == '癸卯'

    def test24(self):
        """二十八宿"""
        star = LunarDay(2020, 4, 13).get_twenty_eight_star()
        assert star.get_zone().get_name() == '南'
        assert star.get_zone().get_beast().get_name() == '朱雀'
        assert star.get_name() == '翼'
        assert star.get_seven_star().get_name() == '火'
        assert star.get_animal().get_name() == '蛇'
        assert star.get_luck().get_name() == '凶'

        assert star.get_land().get_name() == '阳天'
        assert star.get_land().get_direction().get_name() == '东南'

    def test25(self):
        star = LunarDay(2023, 9, 28).get_twenty_eight_star()
        assert star.get_zone().get_name() == '南'
        assert star.get_zone().get_beast().get_name() == '朱雀'
        assert star.get_name() == '柳'
        assert star.get_seven_star().get_name() == '土'
        assert star.get_animal().get_name() == '獐'
        assert star.get_luck().get_name() == '凶'

        assert star.get_land().get_name() == '炎天'
        assert star.get_land().get_direction().get_name() == '南'

    def test26(self):
        lunar: LunarDay = LunarDay(2005, 11, 23)
        assert lunar.get_lunar_month().get_sixty_cycle().get_name() == '戊子'
        assert lunar.get_sixty_cycle_day().get_month().get_name() == '戊子'

    def test27(self):
        lunar: LunarDay = LunarDay(2024, 1, 1)
        assert lunar.next(31).__str__() == '农历甲辰年二月初三'

    def test28(self):
        lunar: LunarDay = LunarDay(2024, 3, 5)
        assert lunar.get_minor_ren().get_name() == '大安'



# -*- coding(self):utf-8 -*-
import unittest

from tyme4py.culture import Phase
from tyme4py.lunar import LunarDay
from tyme4py.solar import SolarDay, SolarTime


class TestPhase(unittest.TestCase):
    
    def test0(self):
        phase = Phase.from_name(2025, 7, '下弦月')
        assert '2025年9月14日 18(self):32(self):57', phase.get_solar_time().__str__()

    def test1(self):
        phase = Phase.from_index(2025, 7, 6)
        assert '2025年9月14日 18(self):32(self):57', phase.get_solar_time().__str__()

    def test2(self):
        phase = Phase.from_index(2025, 7, 8)
        assert '2025年9月22日 03(self):54(self):07', phase.get_solar_time().__str__()

    def test3(self):
        phase = SolarDay.from_ymd(2025, 9, 21).get_phase()
        assert '残月', phase.__str__()

    def test4(self):
        phase = LunarDay.from_ymd(2025, 7, 30).get_phase()
        assert '残月', phase.__str__()

    def test5(self):
        phase = SolarTime.from_ymd_hms(2025, 9, 22, 4, 0, 0).get_phase()
        assert '蛾眉月', phase.__str__()

    def test6(self):
        phase = SolarTime.from_ymd_hms(2025, 9, 22, 3, 0, 0).get_phase()
        assert '残月', phase.__str__()

    def test7(self):
        d = SolarDay.from_ymd(2023, 9, 15).get_phase_day()
        assert '新月第1天', d.__str__()

    def test8(self):
        d = SolarDay.from_ymd(2023, 9, 17).get_phase_day()
        assert '蛾眉月第2天', d.__str__()

    def test9(self):
        phase = SolarTime.from_ymd_hms(2025, 9, 22, 3, 54, 7).get_phase()
        assert '新月', phase.__str__()

    def test10(self):
        phase = SolarTime.from_ymd_hms(2025, 9, 22, 3, 54, 6).get_phase()
        assert '残月', phase.__str__()

    def test11(self):
        phase = SolarTime.from_ymd_hms(2025, 9, 22, 3, 54, 8).get_phase()
        assert '蛾眉月', phase.__str__()



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import List, Union

from tyme4py import LoopTyme
from tyme4py.culture import Luck


class TwelveStar(LoopTyme):
    """黄道黑道十二神"""
    NAMES: List[str] = ['青龙', '明堂', '天刑', '朱雀', '金匮', '天德', '白虎', '玉堂', '天牢', '玄武', '司命', '勾陈']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> TwelveStar:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> TwelveStar:
        return cls(index)

    def next(self, n: int) -> TwelveStar:
        return TwelveStar(self.next_index(n))

    def get_ecliptic(self) -> Ecliptic:
        """
        黄道黑道
        :return: 黄道黑道
        """
        return Ecliptic([0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0, 1][self.get_index()])


class Ecliptic(LoopTyme):
    """黄道黑道就两种，依次为：黄道、黑道。"""
    NAMES: List[str] = ["黄道", "黑道"]
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> Ecliptic:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> Ecliptic:
        return cls(index)

    def next(self, n: int) -> Ecliptic:
        return Ecliptic(self.next_index(n))

    def get_luck(self) -> Luck:
        """
        吉凶
        :return: 吉凶
        """
        return Luck(self.get_index())



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarHalfYear


class TestSolarHalfYear(unittest.TestCase):
    def test0(self):
        assert SolarHalfYear(2023, 0).get_name() == '上半年'
        assert SolarHalfYear(2023, 0).__str__() == '2023年上半年'

    def test1(self):
        assert SolarHalfYear(2023, 1).get_name() == '下半年'
        assert SolarHalfYear(2023, 1).__str__() == '2023年下半年'

    def test2(self):
        assert SolarHalfYear(2023, 0).next(1).get_name() == '下半年'
        assert SolarHalfYear(2023, 0).next(1).__str__() == '2023年下半年'

    def test3(self):
        assert SolarHalfYear(2023, 0).next(2).get_name() == '上半年'
        assert SolarHalfYear(2023, 0).next(2).__str__() == '2024年上半年'

    def test4(self):
        assert SolarHalfYear(2023, 0).next(-2).get_name() == '上半年'
        assert SolarHalfYear(2023, 0).next(-2).__str__() == '2022年上半年'

    def test5(self):
        assert SolarHalfYear(2023, 0).next(-4).__str__() == '2021年上半年'
        assert SolarHalfYear(2023, 0).next(-3).__str__() == '2021年下半年'



# -*- coding:utf-8 -*-


from tyme4py.enums import HideHeavenStemType
from tyme4py.sixtycycle import HideHeavenStemDay
from tyme4py.solar import SolarDay


class TestHideHeavenStem:
    def test0(self):
        d: HideHeavenStemDay = SolarDay(2024, 12, 4).get_hide_heaven_stem_day()
        assert d.get_hide_heaven_stem().get_type() == HideHeavenStemType.MAIN
        assert d.get_hide_heaven_stem().get_name() == '壬'
        assert d.get_hide_heaven_stem().__str__() == '壬'
        assert d.get_hide_heaven_stem().get_heaven_stem().get_element().get_name() == '水'
        assert d.get_name() == '壬水'
        assert d.get_day_index() == 15
        assert d.__str__() == '壬水第16天'

    def test1(self):
        d: HideHeavenStemDay = SolarDay(2024, 11, 7).get_hide_heaven_stem_day()
        assert d.get_hide_heaven_stem().get_type() == HideHeavenStemType.RESIDUAL
        assert d.get_hide_heaven_stem().get_name() == '戊'
        assert d.get_hide_heaven_stem().__str__() == '戊'
        assert d.get_hide_heaven_stem().get_heaven_stem().get_element().get_name() == '土'
        assert d.get_name() == '戊土'
        assert d.get_day_index() == 0
        assert d.__str__() == '戊土第1天'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.festival import SolarFestival
from tyme4py.solar import SolarDay


class TestSolarFestival(unittest.TestCase):
    def test2(self):
        f = SolarFestival.from_index(2023, 0)
        assert f
        f1 = f.next(13)
        assert f1
        assert f1.__str__() == '2024年5月1日 五一劳动节'
        f2 = f.next(-3)
        assert f2
        assert f2.__str__() == '2022年8月1日 八一建军节'

    def test3(self):
        f = SolarFestival.from_index(2023, 0)
        assert f
        f1 = f.next(-9)
        assert f1
        assert f1.__str__() == '2022年3月8日 三八妇女节'

    def test4(self):
        f = SolarDay(2010, 1, 1).get_festival()
        assert f
        assert f.__str__() == '2010年1月1日 元旦'

    def test5(self):
        f = SolarDay(2021, 5, 4).get_festival()
        assert f
        assert f.__str__() == '2021年5月4日 五四青年节'

    def test6(self):
        f = SolarDay(1939, 5, 4).get_festival()
        assert f is None



# -*- coding:utf-8 -*-
from __future__ import annotations

from typing import List, Union

from tyme4py import LoopTyme, AbstractCultureDay


class PlumRain(LoopTyme):
    """梅雨"""
    NAMES: List[str] = ['入梅', '出梅']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> PlumRain:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> PlumRain:
        return cls(index)

    def next(self, n: int) -> PlumRain:
        return PlumRain(self.next_index(n))


class PlumRainDay(AbstractCultureDay):
    """梅雨天"""

    def __init__(self, plum_rain: PlumRain, day_index: int):
        super().__init__(plum_rain, day_index)

    def get_plum_rain(self) -> PlumRain:
        """
        梅雨
        :return: 梅雨
        """
        return super().get_culture()

    def __str__(self) -> str:
        return super().__str__() if self.get_plum_rain().get_index() == 0 else self.get_plum_rain().get_name()



# -*- coding:utf-8 -*-
import unittest

from tyme4py.festival import LunarFestival
from tyme4py.lunar import LunarDay


class TestLunarFestival(unittest.TestCase):
    def test2(self):
        f = LunarFestival.from_index(2023, 0)
        assert f
        assert f.next(13).__str__() == '农历甲辰年正月初一 春节'
        assert f.next(-3).__str__() == '农历壬寅年十一月廿九 冬至节'

    def test3(self):
        f = LunarFestival.from_index(2023, 0)
        assert f
        assert f.next(-9).__str__() == '农历壬寅年三月初五 清明节'

    def test4(self):
        f = LunarDay(2010, 1, 15).get_festival()
        assert f
        assert f.__str__() == '农历庚寅年正月十五 元宵节'

    def test5(self):
        f = LunarDay(2021, 12, 29).get_festival()
        assert f
        assert f.__str__() == '农历辛丑年十二月廿九 除夕'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.solar import SolarDay


class TestFetus(unittest.TestCase):
    def test(self):
        assert SolarDay(2021, 11, 13).get_lunar_day().get_fetus_day().get_name() == '碓磨厕 外东南'

    def test1(self):
        assert SolarDay(2021, 11, 12).get_lunar_day().get_fetus_day().get_name() == '占门碓 外东南'

    def test2(self):
        assert SolarDay(2011, 11, 12).get_lunar_day().get_fetus_day().get_name() == '厨灶厕 外西南'



# -*- coding:utf-8 -*-
import unittest

from tyme4py.lunar import LunarHour


class TestLunarHour(unittest.TestCase):
    def test28(self):
        h: LunarHour = LunarHour(2024, 9, 7, 10, 0, 0)
        assert h.get_minor_ren().get_name() == '留连'



from setuptools import setup

setup(
    name='tyme4py',
    version='1.4.1',
    description='Tyme是一个非常强大的日历工具库，可以看作 Lunar 的升级版，拥有更优的设计和扩展性，支持公历、农历、藏历、星座、干支、生肖、节气、法定假日等。',
    long_description='Tyme是一个非常强大的日历工具库，可以看作 Lunar 的升级版，拥有更优的设计和扩展性，支持公历、农历、藏历、星座、干支、生肖、节气、法定假日等。',
    packages=[
        'tyme4py',
        'tyme4py.culture',
        'tyme4py.culture.dog',
        'tyme4py.culture.fetus',
        'tyme4py.culture.nine',
        'tyme4py.culture.pengzu',
        'tyme4py.culture.phenology',
        'tyme4py.culture.plumrain',
        'tyme4py.culture.ren',
        'tyme4py.culture.star',
        'tyme4py.eightchar',
        'tyme4py.eightchar.provider',
        'tyme4py.eightchar.provider.impl',
        'tyme4py.enums',
        'tyme4py.festival',
        'tyme4py.holiday',
        'tyme4py.jd',
        'tyme4py.lunar',
        'tyme4py.sixtycycle',
        'tyme4py.solar',
        'tyme4py.rabbyung',
        'tyme4py.util'
    ],
    url='https://github.com/6tail/tyme4py',
    license='MIT',
    author='6tail',
    author_email='6tail@6tail.cn',
    python_requires='>=3.8.9',
    keywords='公历 农历 藏历 儒略日 星座 干支 节气 法定假日'
)



# -*- coding:utf-8 -*-
import unittest

from tyme4py.culture import God
from tyme4py.lunar import LunarDay
from tyme4py.solar import SolarDay


class TestGod(unittest.TestCase):
    def test0(self):
        lunar: LunarDay = SolarDay(2004, 2, 16).get_lunar_day()
        gods: [God] = lunar.get_gods()
        ji: [str] = []
        for god in gods:
            if '吉' == god.get_luck().get_name():
                ji.append(god.get_name())
        xiong: [str] = []
        for god in gods:
            if '凶' == god.get_luck().get_name():
                xiong.append(god.get_name())
        assert ji == ['天恩', '续世', '明堂']
        assert xiong == ['月煞', '月虚', '血支', '天贼', '五虚', '土符', '归忌', '血忌']

    def test1(self):
        lunar: LunarDay = SolarDay(2029, 11, 16).get_lunar_day()
        gods: [God] = lunar.get_gods()
        ji: [str] = []
        for god in gods:
            if '吉' == god.get_luck().get_name():
                ji.append(god.get_name())
        xiong: [str] = []
        for god in gods:
            if '凶' == god.get_luck().get_name():
                xiong.append(god.get_name())
        assert ji == ['天德合', '月空', '天恩', '益后', '金匮']
        assert xiong == ['月煞', '月虚', '血支', '五虚']

    def test2(self):
        lunar: LunarDay = SolarDay(1954, 7, 16).get_lunar_day()
        gods: [God] = lunar.get_gods()
        ji: [str] = []
        for god in gods:
            if '吉' == god.get_luck().get_name():
                ji.append(god.get_name())
        xiong: [str] = []
        for god in gods:
            if '凶' == god.get_luck().get_name():
                xiong.append(god.get_name())
        assert ji == ['民日', '天巫', '福德', '天仓', '不将', '续世', '除神', '鸣吠']
        assert xiong == ['劫煞', '天贼', '五虚', '五离']

    def test3(self):
        lunar: LunarDay = SolarDay(2024, 12, 27).get_lunar_day()
        gods: [God] = lunar.get_gods()
        ji: [str] = []
        for god in gods:
            if '吉' == god.get_luck().get_name():
                ji.append(god.get_name())
        xiong: [str] = []
        for god in gods:
            if '凶' == god.get_luck().get_name():
                xiong.append(god.get_name())
        assert ji == ['天恩', '四相', '阴德', '守日', '吉期', '六合', '普护', '宝光']
        assert xiong == ['三丧', '鬼哭']



# -*- coding:utf-8 -*-
from __future__ import annotations

from math import floor, ceil
from typing import TYPE_CHECKING, Union, List

from tyme4py import LoopTyme, AbstractCulture, AbstractCultureDay, AbstractTyme
from tyme4py.culture import Element, Direction, Zodiac, Terrain, Sound, Ten, Twenty, Duty, God, Taboo
from tyme4py.culture.fetus import FetusDay
from tyme4py.culture.pengzu import PengZuEarthBranch, PengZuHeavenStem, PengZu
from tyme4py.culture.star.nine import NineStar
from tyme4py.culture.star.ten import TenStar
from tyme4py.culture.star.twelve import TwelveStar
from tyme4py.eightchar import EightChar
from tyme4py.enums import YinYang, HideHeavenStemType

if TYPE_CHECKING:
    from tyme4py.solar import SolarDay, SolarTime
    from tyme4py.culture.star.twentyeight import TwentyEightStar


class ThreePillars(AbstractCulture):
    """
    三柱
    """
    _year: SixtyCycle
    """年柱"""
    _month: SixtyCycle
    """月柱"""
    _day: SixtyCycle
    """日柱"""

    def __init__(self, year: Union[SixtyCycle, str], month: Union[SixtyCycle, str], day: Union[SixtyCycle, str]):
        """
        :param year: 年干支
        :param month: 月干支
        :param day: 日干支
        """
        self._year = year if isinstance(year, SixtyCycle) else SixtyCycle(year)
        self._month = month if isinstance(month, SixtyCycle) else SixtyCycle(month)
        self._day = day if isinstance(day, SixtyCycle) else SixtyCycle(day)

    def get_year(self) -> SixtyCycle:
        """
        年柱
        :return: 干支 SixtyCycle
        """
        return self._year

    def get_month(self) -> SixtyCycle:
        """
        月柱
        :return: 干支 SixtyCycle
        """
        return self._month

    def get_day(self) -> SixtyCycle:
        """
        日柱
        :return:  干支 SixtyCycle
        """
        return self._day

    def get_name(self) -> str:
        return f'{self._year} {self._month} {self._day}'

    def get_solar_days(self, start_year: int, end_year: int) -> List[SolarDay]:
        """
        三柱转公历日列表
        :param start_year: 开始年份，支持1-9999年
        :param end_year: 结束年份，支持1-9999年
        :return: 公历日 SolarDay的列表
        """
        from tyme4py.solar import SolarDay, SolarTerm
        l: List[SolarDay] = []
        # 月地支距寅月的偏移值
        m: int = self._month.get_earth_branch().next(-2).get_index()
        # 月天干要一致
        if not HeavenStem((self._year.get_heaven_stem().get_index() + 1) * 2 + m) == self._month.get_heaven_stem():
            return l

        # 1年的立春是辛酉，序号57
        y: int = self._year.next(-57).get_index() + 1
        # 节令偏移值
        m *= 2
        base_year: int = start_year - 1
        if base_year > y:
            y += 60 * int(ceil((base_year - y) / 60.0))

        while y <= end_year:
            # 立春为寅月的开始
            term: SolarTerm = SolarTerm(y, 3)
            # 节令推移，年干支和月干支就都匹配上了
            if m > 0:
                term = term.next(m)

            solar_day: SolarDay = term.get_solar_day()
            if solar_day.get_year() >= start_year:
                # 日干支和节令干支的偏移值
                d: int = self._day.next(-solar_day.get_lunar_day().get_sixty_cycle().get_index()).get_index()
                if d > 0:
                    # 从节令推移天数
                    solar_day = solar_day.next(d)

                # 验证一下
                if solar_day.get_sixty_cycle_day().get_three_pillars() == self:
                    l.append(solar_day)
            y += 60
        return l


class EarthBranch(LoopTyme):
    """地支（地元）"""
    NAMES: List[str] = ['子', '丑', '寅', '卯', '辰', '巳', '午', '未', '申', '酉', '戌', '亥']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> EarthBranch:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> EarthBranch:
        return cls(index)

    def next(self, n: int) -> EarthBranch:
        return EarthBranch(self.next_index(n))

    def get_element(self) -> Element:
        """
        :return:五行
        """
        return Element([4, 2, 0, 0, 2, 1, 1, 2, 3, 3, 2, 4][self.get_index()])

    def get_yin_yang(self) -> YinYang:
        """
        :return:阴阳
        """
        return YinYang.YANG if self.get_index() % 2 == 0 else YinYang.YIN

    def get_hide_heaven_stem_main(self) -> HeavenStem:
        """
        藏干之本气(主气)
        :return: 天干 HeavenStem
        """
        return HeavenStem([9, 5, 0, 1, 4, 2, 3, 5, 6, 7, 4, 8][self.get_index()])

    def get_hide_heaven_stem_middle(self) -> Union[HeavenStem, None]:
        """
        藏干之中气，无中气的返回None
        :return: 天干 HeavenStem
        """
        n: int = [-1, 9, 2, -1, 1, 6, 5, 3, 8, -1, 7, 0][self.get_index()]
        return None if n == -1 else HeavenStem(n)

    def get_hide_heaven_stem_residual(self) -> Union[HeavenStem, None]:
        """
        藏干之余气，无余气的返回None
        :return: 天干 HeavenStem
        """
        n: int = [-1, 7, 4, -1, 9, 4, -1, 1, 4, -1, 3, -1][self.get_index()]
        return None if n == -1 else HeavenStem(n)

    def get_hide_heaven_stems(self) -> List[HideHeavenStem]:
        """
        :return: 藏干列表
        """
        l: List[HideHeavenStem] = [HideHeavenStem(self.get_hide_heaven_stem_main(), HideHeavenStemType.MAIN)]
        o: Union[HeavenStem, None] = self.get_hide_heaven_stem_middle()
        if o:
            l.append(HideHeavenStem(o, HideHeavenStemType.MIDDLE))
        o = self.get_hide_heaven_stem_residual()
        if o:
            l.append(HideHeavenStem(o, HideHeavenStemType.RESIDUAL))
        return l

    def get_zodiac(self) -> Zodiac:
        """
        生肖属相
        :return:生肖 Zodiac
        """
        return Zodiac(self.get_index())

    def get_direction(self) -> Direction:
        """
        :return: 方位 Direction
        """
        return Direction([0, 4, 2, 2, 4, 8, 8, 4, 6, 6, 4, 0][self.get_index()])

    def get_opposite(self) -> EarthBranch:
        """
        六冲
        子午冲，丑未冲，寅申冲，辰戌冲，卯酉冲，巳亥冲。
        :return: 地支 EarthBranch
        """
        return self.next(6)

    def get_ominous(self) -> Direction:
        """
        煞
        逢巳日、酉日、丑日必煞东；
        亥日、卯日、未日必煞西；
        申日、子日、辰日必煞南；
        寅日、午日、戌日必煞北。
        :return: 方位 Direction
        """
        return Direction([8, 2, 0, 6][self.get_index() % 4])

    def get_peng_zu_earth_branch(self) -> PengZuEarthBranch:
        """
        :return: 地支彭祖百忌 PengZuEarthBranch
        """
        return PengZuEarthBranch(self.get_index())

    def get_combine(self) -> EarthBranch:
        """
        六合（子丑合，寅亥合，卯戌合，辰酉合，巳申合，午未合）
        :return: 地支 EarthBranch
        """
        return EarthBranch(1 - self.get_index())

    def get_harm(self) -> EarthBranch:
        """
        六害（子未害、丑午害、寅巳害、卯辰害、申亥害、酉戌害）
        :return: 地支 EarthBranch
        """
        return EarthBranch(19 - self.get_index())

    def combine(self, target: EarthBranch) -> Union[Element, None]:
        """
        合化
        子丑合化土，寅亥合化木，卯戌合化火，辰酉合化金，巳申合化水，午未合化火
        :param target: 地支 EarthBranch
        :return: 能合化则返回五行属性，不能合化则返回None
        """
        return Element([2, 2, 0, 1, 3, 4, 2, 2, 4, 3, 1, 0][self.get_index()]) if self.get_combine() == target else None


class HeavenStem(LoopTyme):
    """天干（天元）"""
    NAMES: List[str] = ['甲', '乙', '丙', '丁', '戊', '己', '庚', '辛', '壬', '癸']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self.NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> HeavenStem:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> HeavenStem:
        return cls(index)

    def next(self, n: int) -> HeavenStem:
        return HeavenStem(self.next_index(n))

    def get_element(self) -> Element:
        """
        五行
        :return: 五行
        """
        return Element(self.get_index() // 2)

    def get_yin_yang(self) -> YinYang:
        """
        :return: 阴阳
        """
        return YinYang.YANG if self.get_index() % 2 == 0 else YinYang.YIN

    def get_ten_star(self, target: HeavenStem) -> TenStar:
        """
        十神（生我者，正印偏印。我生者，伤官食神。克我者，正官七杀。我克者，正财偏财。同我者，劫财比肩。）
        :param target: 天干
        :return:十神
        """
        target_index: int = target.get_index()
        offset: int = target_index - self.get_index()
        if self.get_index() % 2 != 0 and target_index % 2 == 0:
            offset += 2
        return TenStar(offset)

    def get_direction(self) -> Direction:
        """
        :return: 方位
        """
        return self.get_element().get_direction()

    def get_joy_direction(self) -> Direction:
        """
        喜神方位（《喜神方位歌》甲己在艮乙庚乾，丙辛坤位喜神安。丁壬只在离宫坐，戊癸原在在巽间。）
        :return: 方位
        """
        return Direction([7, 5, 1, 8, 3][self.get_index() % 5])

    def get_yang_direction(self) -> Direction:
        """
        阳贵神方位（《阳贵神歌》甲戊坤艮位，乙己是坤坎，庚辛居离艮，丙丁兑与乾，震巽属何日，壬癸贵神安。）
        :return: 方位
        """
        return Direction([1, 1, 6, 5, 7, 0, 8, 7, 2, 3][self.get_index()])

    def get_yin_direction(self) -> Direction:
        """
        阴贵神方位（《阴贵神歌》甲戊见牛羊，乙己鼠猴乡，丙丁猪鸡位，壬癸蛇兔藏，庚辛逢虎马，此是贵神方。）
        :return: 方位
        """
        return Direction([7, 0, 5, 6, 1, 1, 7, 8, 3, 2][self.get_index()])

    def get_wealth_direction(self) -> Direction:
        """
        财神方位（《财神方位歌》甲乙东北是财神，丙丁向在西南寻，戊己正北坐方位，庚辛正东去安身，壬癸原来正南坐，便是财神方位真。）
        :return:方位
        """
        return Direction([7, 1, 0, 2, 8][self.get_index() // 2])

    def get_mascot_direction(self) -> Direction:
        """
        福神方位（《福神方位歌》甲乙东南是福神，丙丁正东是堪宜，戊北己南庚辛坤，壬在乾方癸在西。）
        :return:方位
        """
        return Direction([3, 3, 2, 2, 0, 8, 1, 1, 5, 6][self.get_index()])

    def get_peng_zu_heaven_stem(self) -> PengZuHeavenStem:
        """
        :return:天干彭祖百忌
        """
        return PengZuHeavenStem(self.get_index())

    def get_terrain(self, earth_branch: EarthBranch) -> Terrain:
        """
        长生十二神(地势)
        长生十二神可通过不同的组合，得到自坐和星运。
        :param: 地支
        :return：地势(长生十二神)
        """
        earth_branch_index: int = earth_branch.get_index()
        return Terrain([1, 6, 10, 9, 10, 9, 7, 0, 4, 3][self.get_index()] + (earth_branch_index if YinYang.YANG == self.get_yin_yang() else -earth_branch_index))

    def get_combine(self) -> HeavenStem:
        """
        天干五合（甲己合，乙庚合，丙辛合，丁壬合，戊癸合）
        :return: 天干
        """
        return self.next(5)

    def combine(self, target: HeavenStem) -> Element:
        """
        合化（甲己合化土 乙庚合化金 丙辛合化水 丁壬合化木 戊癸合化火）
        :param target: 天干
        :return: 能合化则返回五行属性，不能合化则返回null
        """
        return Element(self.get_index() + 2) if self.get_combine() == target else None


class HideHeavenStem(AbstractCulture):
    """藏干（即人元，司令取天干，分野取天干的五行）"""
    _heaven_stem: HeavenStem
    """天干"""
    _type: HideHeavenStemType
    """藏干类型"""

    def __init__(self, heaven_stem: Union[HeavenStem, str, int], hide_heaven_stem_type: HideHeavenStemType):
        if isinstance(heaven_stem, int) or isinstance(heaven_stem, str):
            self._heaven_stem = HeavenStem(heaven_stem)
        else:
            self._heaven_stem = heaven_stem
        self._type = hide_heaven_stem_type

    def get_name(self) -> str:
        return self._heaven_stem.get_name()

    def get_heaven_stem(self) -> HeavenStem:
        """
        天干
        :return: 天干
        """
        return self._heaven_stem

    def get_type(self) -> HideHeavenStemType:
        """
        藏干类型
        :return: 藏干类型
        """
        return self._type


class HideHeavenStemDay(AbstractCultureDay):
    """人元司令分野（地支藏干+天索引）"""

    def __init__(self, hide_heaven_stem: HideHeavenStem, day_index: int):
        super().__init__(hide_heaven_stem, day_index)

    def get_hide_heaven_stem(self) -> HideHeavenStem:
        """
        :return: 藏干
        """
        return self.get_culture()

    def get_name(self) -> str:
        heaven_stem: HeavenStem = self.get_hide_heaven_stem().get_heaven_stem()
        return f'{heaven_stem.get_name()}{heaven_stem.get_element().get_name()}'


class SixtyCycle(LoopTyme):
    """六十甲子(六十干支周)"""
    _NAMES: List[str] = ['甲子', '乙丑', '丙寅', '丁卯', '戊辰', '己巳', '庚午', '辛未', '壬申', '癸酉', '甲戌', '乙亥', '丙子', '丁丑', '戊寅', '己卯', '庚辰', '辛巳', '壬午', '癸未', '甲申', '乙酉', '丙戌', '丁亥', '戊子', '己丑', '庚寅', '辛卯', '壬辰', '癸巳', '甲午', '乙未', '丙申', '丁酉', '戊戌', '己亥', '庚子', '辛丑', '壬寅', '癸卯', '甲辰', '乙巳', '丙午', '丁未', '戊申', '己酉', '庚戌', '辛亥', '壬子', '癸丑', '甲寅', '乙卯', '丙辰', '丁巳', '戊午', '己未', '庚申', '辛酉', '壬戌', '癸亥']
    """名称"""

    def __init__(self, index_or_name: Union[int, str]):
        super().__init__(self._NAMES, index_or_name)

    @classmethod
    def from_name(cls, name: str) -> SixtyCycle:
        return cls(name)

    @classmethod
    def from_index(cls, index: int) -> SixtyCycle:
        return cls(index)

    def next(self, n: int) -> SixtyCycle:
        return SixtyCycle(self.next_index(n))

    def get_heaven_stem(self) -> HeavenStem:
        """
        天干
        :return: 天干
        """
        return HeavenStem(self.get_index() % 10)

    def get_earth_branch(self) -> EarthBranch:
        """
        地支
        :return: 地支
        """
        return EarthBranch(self.get_index() % 12)

    def get_sound(self) -> Sound:
        """
        纳音
        :return: 纳音
        """
        return Sound(self.get_index() // 2)

    def get_peng_zu(self) -> PengZu:
        """
        彭祖百忌
        :return: 彭祖百忌
        """
        return PengZu(self)

    def get_ten(self) -> Ten:
        """
        旬
        :return: 旬
        """
        return Ten((self.get_heaven_stem().get_index() - self.get_earth_branch().get_index()) // 2)

    def get_extra_earth_branches(self) -> List[EarthBranch]:
        """
        旬空(空亡)，因地支比天干多2个，旬空则为每一轮干支一一配对后多出来的2个地支
        :return: 旬空(空亡)
        """
        earth_branch: EarthBranch = EarthBranch(10 + self.get_earth_branch().get_index() - self.get_heaven_stem().get_index())
        return [earth_branch, earth_branch.next(1)]


class SixtyCycleYear(AbstractTyme):
    """
    干支年
    """
    _year: int

    def __init__(self, year: int):
        if year < -1 or year > 9999:
            raise ValueError(f'illegal sixty cycle year: {year}')
        self._year = year

    @classmethod
    def from_year(cls, year: int) -> SixtyCycleYear:
        return cls(year)

    def get_year(self) -> int:
        """
        年
        :return: 返回为干支年数字，范围为-1到9999。
        """
        return self._year

    def get_name(self) -> str:
        """
        :return: {六十甲子}年
        """
        return f'{self.get_sixty_cycle().get_name()}年'

    def next(self, n: int) -> SixtyCycleYear:
        return SixtyCycleYear(self._year + n)

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        干支
        :return: 干支 SixtyCycle
        """
        return SixtyCycle(self._year - 4)

    def get_twenty(self) -> Twenty:
        """
        运
        :return: 返回为运 Twenty。
        """
        return Twenty(floor((self._year - 1864) / 20))

    def get_nine_star(self) -> NineStar:
        """
        九星
        :return: 返回为九星 NineStar。
        """
        return NineStar(63 + self.get_twenty().get_sixty().get_index() * 3 - self.get_sixty_cycle().get_index())

    def get_jupiter_direction(self) -> Direction:
        """
        太岁方位
        :return: 返回为方位 Direction。
        """
        return Direction([0, 7, 7, 2, 3, 3, 8, 1, 1, 6, 0, 0][self.get_sixty_cycle().get_earth_branch().get_index()])

    def get_first_month(self) -> SixtyCycleMonth:
        """
        首月（依据五虎遁和正月起寅的规律）
        :return: 干支月
        """
        h: HeavenStem = HeavenStem.from_index((self.get_sixty_cycle().get_heaven_stem().get_index() + 1) * 2)
        return SixtyCycleMonth(self, SixtyCycle.from_name(f'{h.get_name()}寅'))

    def get_months(self) -> List[SixtyCycleMonth]:
        """
        干支月列表
        :return: 干支月列表
        """
        l: List[SixtyCycleMonth] = []
        m: SixtyCycleMonth = self.get_first_month()
        l.append(m)
        for i in range(1, 12):
            l.append(m.next(i))
        return l


class SixtyCycleMonth(AbstractTyme):
    """
    干支月
    """
    _year: SixtyCycleYear
    """干支年"""
    _month: SixtyCycle
    """月柱"""

    def __init__(self, year: SixtyCycleYear, month: SixtyCycle):
        """
        :param year: 干支年
        :param month: 干支月
        """
        self._year = year
        self._month = month

    @classmethod
    def from_index(cls, year: int, index: int) -> SixtyCycleMonth:
        return SixtyCycleYear.from_year(year).get_first_month().next(index)

    def get_sixty_cycle_year(self) -> SixtyCycleYear:
        """
        干支年
        :return: 干支年
        """
        return self._year

    def get_year(self) -> SixtyCycle:
        """
        年柱
        :return: 干支
        """
        return self._year.get_sixty_cycle()

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        干支
        :return: 干支
        """
        return self._month

    def get_index_in_year(self) -> int:
        """
        位于当年的索引(0-11)，寅月为0，依次类推
        :return: 返回为数字，范围0到11，寅月为0，依次类推。
        """
        return self._month.get_earth_branch().next(-2).get_index()

    def get_name(self) -> str:
        """
        名称
        :return: 名称
        """
        return f'{self._month}月'

    def __str__(self) -> str:
        return f'{self._year}{self.get_name()}'

    def next(self, n: int) -> SixtyCycleMonth:
        return SixtyCycleMonth(SixtyCycleYear.from_year((self._year.get_year()*12 + self.get_index_in_year() + n) // 12), self._month.next(n))

    def get_first_day(self) -> SixtyCycleDay:
        from tyme4py.solar import SolarTerm
        return SixtyCycleDay.from_solar_day(SolarTerm.from_index(self._year.get_year(), 3+self.get_index_in_year()*2).get_solar_day())

    def get_days(self) -> List[SixtyCycleDay]:
        """
        本月的干支日列表
        :return: 支日列表
        """
        l: List[SixtyCycleDay] = []
        d: SixtyCycleDay = self.get_first_day()
        while d.get_sixty_cycle_month() == self:
            l.append(d)
            d = d.next(1)
        return l

    def get_nine_star(self) -> NineStar:
        """
        :return: 九星 NineStar。
        """
        index = self._month.get_earth_branch().get_index()
        if index < 2:
            index += 3
        return NineStar(27 - self.get_year().get_earth_branch().get_index() % 3 * 3 - index)

    def get_jupiter_direction(self) -> Direction:
        """
        太岁方位
        :return: 方位 Direction。
        """
        n: int = [7, -1, 1, 3][self._month.get_earth_branch().next(-2).get_index() % 4]
        return self._month.get_heaven_stem().get_direction() if n == -1 else Direction(n)


class SixtyCycleDay(AbstractTyme):
    """干支日"""
    _solar_day: SolarDay
    """公历日"""
    _month: SixtyCycleMonth
    """干支月"""
    _day: SixtyCycle
    """日柱"""

    def __init__(self, solar_day: SolarDay, month: SixtyCycleMonth, day: SixtyCycle):
        """
        :param solar_day: 公历年
        :param month: 干支月
        :param day: 日柱
        """
        self._solar_day = solar_day
        self._month = month
        self._day = day

    @classmethod
    def from_solar_day(cls, solar_day: SolarDay) -> SixtyCycleDay:
        from tyme4py.solar import SolarTerm, SolarDay
        from tyme4py.lunar import LunarYear, LunarMonth, LunarDay
        solar_year: int = solar_day.get_year()
        spring_solar_day: SolarDay = SolarTerm.from_index(solar_year, 3).get_solar_day()
        lunar_day: LunarDay = solar_day.get_lunar_day()
        lunar_year: LunarYear = lunar_day.get_lunar_month().get_lunar_year()
        if lunar_year.get_year() == solar_year:
            if solar_day.is_before(spring_solar_day):
                lunar_year = lunar_year.next(-1)
        elif lunar_year.get_year() < solar_year:
            if not solar_day.is_before(spring_solar_day):
                lunar_year = lunar_year.next(1)
        term: SolarTerm = solar_day.get_term()
        index: int = term.get_index() - 3
        if index < 0 and term.get_solar_day().is_after(spring_solar_day):
            index += 24
        return cls(solar_day, SixtyCycleMonth(SixtyCycleYear.from_year(lunar_year.get_year()), LunarMonth.from_ym(solar_year, 1).get_sixty_cycle().next(int(floor(index * 0.5)))), lunar_day.get_sixty_cycle())

    def get_solar_day(self) -> SolarDay:
        """
        :return: 公历日
        """
        return self._solar_day

    def get_sixty_cycle_month(self) -> SixtyCycleMonth:
        """
        :return: 干支月
        """
        return self._month

    def get_year(self) -> SixtyCycle:
        """
        年柱
        :return: 干支
        """
        return self._month.get_year()

    def get_month(self) -> SixtyCycle:
        """
        月柱
        :return: 干支
        """
        return self._month.get_sixty_cycle()

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        干支
        :return: 干支
        """
        return self._day

    def get_name(self) -> str:
        return f'{self._day}日'

    def __str__(self) -> str:
        return f'{self._month}{self.get_name()}'

    def next(self, n: int) -> SixtyCycleDay:
        return SixtyCycleDay.from_solar_day(self._solar_day.next(n))

    def get_duty(self) -> Duty:
        """
        建除十二值神
        :return: 建除十二值神 Duty。
        """
        return Duty(self._day.get_earth_branch().get_index() - self.get_month().get_earth_branch().get_index())

    def get_twelve_star(self) -> TwelveStar:
        """
        黄道黑道十二神
        :return: 黄道黑道十二神 TwelveStar。
        """
        return TwelveStar(self._day.get_earth_branch().get_index() + (8 - self.get_month().get_earth_branch().get_index() % 6) * 2)

    def get_nine_star(self) -> NineStar:
        """
        九星
        :return: 九星 NineStar。
        """
        from tyme4py.solar import SolarDay, SolarTerm
        d: SolarDay = self._solar_day
        dong_zhi: SolarTerm = SolarTerm(d.get_year(), 0)
        dong_zhi_solar: SolarDay = dong_zhi.get_solar_day()
        xia_zhi_solar: SolarDay = dong_zhi.next(12).get_solar_day()
        dong_zhi_solar2: SolarDay = dong_zhi.next(24).get_solar_day()
        dong_zhi_index: int = dong_zhi_solar.get_lunar_day().get_sixty_cycle().get_index()
        xia_zhi_index: int = xia_zhi_solar.get_lunar_day().get_sixty_cycle().get_index()
        dong_zhi_index2: int = dong_zhi_solar2.get_lunar_day().get_sixty_cycle().get_index()
        solar_shun_bai: SolarDay = dong_zhi_solar.next(60 - dong_zhi_index if dong_zhi_index > 29 else -dong_zhi_index)
        solar_shun_bai2: SolarDay = dong_zhi_solar2.next(60 - dong_zhi_index2 if dong_zhi_index2 > 29 else -dong_zhi_index2)
        solar_ni_zi: SolarDay = xia_zhi_solar.next(60 - xia_zhi_index if xia_zhi_index > 29 else -xia_zhi_index)
        offset: int = 0
        if not d.is_before(solar_shun_bai) and d.is_before(solar_ni_zi):
            offset = d.subtract(solar_shun_bai)
        elif not d.is_before(solar_ni_zi) and d.is_before(solar_shun_bai2):
            offset = 8 - d.subtract(solar_ni_zi)
        elif not d.is_before(solar_shun_bai2):
            offset = d.subtract(solar_shun_bai2)
        elif d.is_before(solar_shun_bai):
            offset = 8 + solar_shun_bai.subtract(d)
        return NineStar(offset)

    def get_jupiter_direction(self) -> Direction:
        """
        太岁方位
        :return: 方位 Direction。
        """
        index: int = self._day.get_index()
        from tyme4py.culture import Element
        return Element(index // 12).get_direction() if index % 12 < 6 else self._month.get_sixty_cycle_year().get_jupiter_direction()

    def get_fetus_day(self) -> FetusDay:
        """
        逐日胎神
        :return:逐日胎神 FetusDay。
        """
        return FetusDay.from_sixty_cycle_day(self)

    def get_twenty_eight_star(self) -> TwentyEightStar:
        """
        二十八宿
        :return: 二十八宿 TwentyEightStar。
        """
        from tyme4py.culture.star.twentyeight import TwentyEightStar
        return TwentyEightStar([10, 18, 26, 6, 14, 22, 2][self._solar_day.get_week().get_index()]).next(-7 * self._day.get_earth_branch().get_index())

    def get_gods(self) -> List[God]:
        """
        神煞列表(吉神宜趋，凶神宜忌)
        :return: 神煞列表
        """
        return God.get_day_gods(self.get_month(), self._day)

    def get_recommends(self) -> List[Taboo]:
        """
        今日 宜
        :return: 宜忌列表
        """
        return Taboo.get_day_recommends(self.get_month(), self._day)

    def get_avoids(self) -> List[Taboo]:
        """
        今日 忌
        :return: 宜忌列表
        """
        return Taboo.get_day_avoids(self.get_month(), self._day)

    def get_hours(self) -> List[SixtyCycleHour]:
        """
        当天的干支时辰列表
        :return: 干支时辰列表
        """
        from tyme4py.solar import SolarDay, SolarTime
        l: List[SixtyCycleHour] = []
        d: SolarDay = self._solar_day.next(-1)
        t: SolarTime = SolarTime.from_ymd_hms(d.get_year(), d.get_month(), d.get_day(), 23, 0, 0)
        h: SixtyCycleHour = SixtyCycleHour.from_solar_time(t)
        l.append(h)
        for i in range(0, 11):
            h = h.next(7200)
            l.append(h)
        return l

    def get_three_pillars(self) -> ThreePillars:
        return ThreePillars(self.get_year(), self.get_month(), self.get_sixty_cycle())


class SixtyCycleHour(AbstractTyme):
    """
    干支时辰
    """
    _solar_time: SolarTime
    """公历时刻"""
    _day: SixtyCycleDay
    """干支日"""
    _hour: SixtyCycle
    """时柱"""

    def __init__(self, solar_time: SolarTime):
        from tyme4py.solar import SolarTerm, SolarTime
        from tyme4py.lunar import LunarYear, LunarMonth, LunarDay, LunarHour
        solar_year: int = solar_time.get_year()
        spring_solar_time: SolarTime = SolarTerm.from_index(solar_year, 3).get_julian_day().get_solar_time()
        lunar_hour: LunarHour = solar_time.get_lunar_hour()
        lunar_day: LunarDay = lunar_hour.get_lunar_day()
        lunar_year: LunarYear = lunar_day.get_lunar_month().get_lunar_year()
        if lunar_year.get_year() == solar_year:
            if solar_time.is_before(spring_solar_time):
                lunar_year = lunar_year.next(-1)
        elif lunar_year.get_year() < solar_year:
            if not solar_time.is_before(spring_solar_time):
                lunar_year = lunar_year.next(1)

        term: SolarTerm = solar_time.get_term()
        index: int = term.get_index() - 3
        if index < 0 and term.get_julian_day().get_solar_time().is_after(SolarTerm.from_index(solar_year, 3).get_julian_day().get_solar_time()):
            index += 24
        d: SixtyCycle = lunar_day.get_sixty_cycle()
        if solar_time.get_hour() == 23:
            d = d.next(1)
        y: SixtyCycleYear = SixtyCycleYear.from_year(lunar_year.get_year())
        m: LunarMonth = LunarMonth.from_ym(solar_year, 1)

        self._solar_time = solar_time
        self._day = SixtyCycleDay(solar_time.get_solar_day(), SixtyCycleMonth(y, m.get_sixty_cycle().next(int(floor(index * 0.5)))), d)
        self._hour = lunar_hour.get_sixty_cycle()

    @classmethod
    def from_solar_time(cls, solar_time: SolarTime) -> SixtyCycleHour:
        return cls(solar_time)

    def get_year(self) -> SixtyCycle:
        """
        年柱
        :return: 干支
        """
        return self._day.get_year()

    def get_month(self) -> SixtyCycle:
        """
        月柱
        :return: 干支
        """
        return self._day.get_month()

    def get_day(self) -> SixtyCycle:
        """
        日柱
        :return: 干支
        """
        return self._day.get_sixty_cycle()

    def get_sixty_cycle(self) -> SixtyCycle:
        """
        干支
        :return: 干支
        """
        return self._hour

    def get_sixty_cycle_day(self) -> SixtyCycleDay:
        """
        干支日
        :return: 干支日
        """
        return self._day

    def get_solar_time(self) -> SolarTime:
        """
        公历时刻
        :return: 公历时刻
        """
        return self._solar_time

    def get_name(self) -> str:
        return f'{self._hour}时'

    def __str__(self) -> str:
        return f'{self._day}{self.get_name()}'

    def get_index_in_day(self) -> int:
        """
        位于当天的索引
        :return: 位于当天的索引
        """
        h: int = self._solar_time.get_hour()
        if h == 23:
            return 0
        return (h + 1) // 2

    def next(self, n: int) -> SixtyCycleHour:
        return SixtyCycleHour.from_solar_time(self._solar_time.next(n))

    def get_twelve_star(self) -> TwelveStar:
        """
        黄道黑道十二神
        :return: 黄道黑道十二神 TwelveStar。
        """
        return TwelveStar(self._hour.get_earth_branch().get_index() + (8 - self.get_day().get_earth_branch().get_index() % 6) * 2)

    def get_nine_star(self) -> NineStar:
        """
        九星（时家紫白星歌诀：三元时白最为佳，冬至阳生顺莫差，孟日七宫仲一白，季日四绿发萌芽，每把时辰起甲子，本时星耀照光华，时星移入中宫去，顺飞八方逐细查。夏至阴生逆回首，孟归三碧季加六，仲在九宫时起甲，依然掌中逆轮跨。）
        :return: 九星 NineStar。
        """
        from tyme4py.solar import SolarTerm, SolarDay
        solar: SolarDay = self._solar_time.get_solar_day()
        dong_zhi: SolarTerm = SolarTerm(solar.get_year(), 0)
        earth_branch_index: int = self.get_index_in_day() % 12
        index: int = [8, 5, 2][self.get_day().get_earth_branch().get_index() % 3]
        if (not solar.is_before(dong_zhi.get_julian_day().get_solar_day())) and solar.is_before(dong_zhi.next(12).get_julian_day().get_solar_day()):
            index = 8 + earth_branch_index - index
        else:
            index -= earth_branch_index
        return NineStar(index)

    def get_eight_char(self) -> EightChar:
        """
        八字
        :return: 八字
        """
        return EightChar(self.get_year(), self.get_month(), self.get_day(), self._hour)

    def get_recommends(self) -> List[Taboo]:
        """
        时辰 宜
        :return: 宜忌列表
        """
        return Taboo.get_hour_recommends(self.get_day(), self._hour)

    def get_avoids(self) -> List[Taboo]:
        """
        时辰 忌
        :return: 宜忌列表
        """
        return Taboo.get_hour_avoids(self.get_day(), self._hour)
